from typing import NoReturn
import numpy
import scipy
from scipy.interpolate import interp2d
from scipy.optimize import minimize

import matplotlib
import matplotlib.pyplot as plt
import matplotlib.widgets as wdg

from cryspy.A_functions_base.powder_diffraction_const_wavelength import \
    calc_ttheta_phi_by_gamma_nu

from cryspy.A_functions_base.function_1_gamma_nu import \
    recal_int_to_gammanu_grid

from cryspy.B_parent_classes.cl_1_item import ItemN

from .cl_1_pd2d_background import Pd2dBackground

na = numpy.newaxis

class Pd2dProc(ItemN):
    """
    Calculated 2D diffraction pattern (polarized neutrons).

    Attributes
    ----------
        - gamma_nu_intensity_plus_net, gamma_nu_intensity_minus_net,
          gamma_nu_intensity_plus_total, gamma_nu_intensity_minus_total,
          gamma_nu_intensity_bkg_calc, gamma_nu_intensity_plus,
          gamma_nu_intensity_plus_sigma, gamma_nu_intensity_minus,
          gamma_nu_intensity_minus_sigma (mandatory)

    Internal
    --------
        - gamma, nu, intensity_plus_net, intensity_minus_net,
          intensity_plus_total, intensity_minus_total, intensity_bkg_calc,
          intensity_plus, intensity_plus_sigma, intensity_minus,
          intensity_minus_sigma

    """
    ATTR_MANDATORY_NAMES = (
        "gamma_nu_intensity_plus_net", "gamma_nu_intensity_minus_net",
        "gamma_nu_intensity_bkg_calc", "gamma_nu_excluded_points", )
    ATTR_MANDATORY_TYPES = (str, str, str, str, )

    ATTR_MANDATORY_CIF = (
        "gamma_nu_intensity_plus_net", "gamma_nu_intensity_minus_net",
        "gamma_nu_intensity_bkg_calc", "gamma_nu_excluded_points", )

    ATTR_OPTIONAL_NAMES = (
        "gamma_nu_intensity_plus", "gamma_nu_intensity_plus_sigma",
        "gamma_nu_intensity_minus", "gamma_nu_intensity_minus_sigma",
        "gamma_nu_intensity", "gamma_nu_intensity_sigma", )
    ATTR_OPTIONAL_TYPES = (str, str, str, str, str, str,  )
    ATTR_OPTIONAL_CIF = (
        "gamma_nu_intensity_plus", "gamma_nu_intensity_plus_sigma",
        "gamma_nu_intensity_minus", "gamma_nu_intensity_minus_sigma",
        "gamma_nu_intensity", "gamma_nu_intensity_sigma", )

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = (
        "gamma", "nu", "intensity_plus_net", "intensity_minus_net",
        "intensity_bkg_calc",
        "intensity_plus", "intensity_plus_sigma",
        "intensity_minus", "intensity_minus_sigma",
        "intensity", "intensity_sigma",
        "excluded_points")
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "pd2d_proc"

    def __init__(self, **kwargs) -> NoReturn:
        super(Pd2dProc, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self) -> NoReturn:
        flag = True
        if not(self.is_attribute("gamma_nu_intensity_plus_net")):
            return False
        l_1 = (self.gamma_nu_intensity_plus_net).strip().split("\n")
        l_2 = (self.gamma_nu_intensity_minus_net).strip().split("\n")
        l_3 = (self.gamma_nu_intensity_bkg_calc).strip().split("\n")
        l_8 = (self.gamma_nu_excluded_points).strip().split("\n")

        flag_polarized, flag_unpolarized = False, False
        if self.is_attribute("gamma_nu_intensity_plus"):
            l_4 = (self.gamma_nu_intensity_plus).strip().split("\n")
            l_5 = (self.gamma_nu_intensity_plus_sigma).strip().split("\n")
            l_6 = (self.gamma_nu_intensity_minus).strip().split("\n")
            l_7 = (self.gamma_nu_intensity_minus_sigma).strip().split("\n")
            flag_polarized = True
        elif self.is_attribute("gamma_nu_intensity"):
            l_4 = (self.gamma_nu_intensity).strip().split("\n")
            l_5 = (self.gamma_nu_intensity_sigma).strip().split("\n")
            flag_unpolarized = True
        else:
            return False
        l_gamma = numpy.array([_ for _ in l_1[0].strip().split()[1:]], dtype=float)
        l_nu = []
        ll_intensity_plus_net, ll_intensity_minus_net = [], []
        ll_excluded_points = []
        ll_intensity_bkg_calc = []
        for line_1, line_2, line_3, line_8 in zip(l_1[1:], l_2[1:], l_3[1:], l_8[1:]):
            _l_1 = line_1.strip().split()
            _l_2 = line_2.strip().split()
            _l_3 = line_3.strip().split()
            _l_8 = line_8.strip().split()
            l_nu.append(float(_l_1[0]))
            ll_intensity_plus_net.append(_l_1[1:])
            ll_intensity_minus_net.append(_l_2[1:])
            ll_intensity_bkg_calc.append(_l_3[1:])
            ll_excluded_points.append(_l_8[1:])

        if flag_polarized:
            ll_intensity_plus, ll_intensity_plus_sigma = [], []
            ll_intensity_minus, ll_intensity_minus_sigma = [], []
            for line_4, line_5, line_6, line_7 in zip(l_4[1:], l_5[1:], l_6[1:], l_7[1:]):
                _l_4 = line_4.strip().split()
                _l_5 = line_5.strip().split()
                _l_6 = line_6.strip().split()
                _l_7 = line_7.strip().split()
                ll_intensity_plus.append(_l_4[1:])
                ll_intensity_plus_sigma.append(_l_5[1:])
                ll_intensity_minus.append(_l_6[1:])
                ll_intensity_minus_sigma.append(_l_7[1:])
        elif flag_unpolarized:
            ll_intensity, ll_intensity_sigma = [], []
            for line_4, line_5 in zip(l_4[1:], l_5[1:]):
                _l_4 = line_4.strip().split()
                _l_5 = line_5.strip().split()
                ll_intensity.append(_l_4[1:])
                ll_intensity_sigma.append(_l_5[1:])

        ll_intensity_plus_net = numpy.array(ll_intensity_plus_net, dtype=float).transpose()
        ll_intensity_minus_net = numpy.array(ll_intensity_minus_net, dtype=float).transpose()
        ll_intensity_bkg_calc = numpy.array(ll_intensity_bkg_calc, dtype=float).transpose()
        ll_excluded_points = numpy.array([[hhh == "True" for hhh in hh ]for hh in ll_excluded_points], dtype=bool).transpose()

        if flag_polarized:
            ll_intensity_plus = numpy.array(ll_intensity_plus, dtype=float).transpose()
            ll_intensity_plus_sigma = numpy.array(ll_intensity_plus_sigma, dtype=float).transpose()
            ll_intensity_minus = numpy.array(ll_intensity_minus, dtype=float).transpose()
            ll_intensity_minus_sigma = numpy.array(ll_intensity_minus_sigma, dtype=float).transpose()
        elif flag_unpolarized:
            ll_intensity = numpy.array(ll_intensity, dtype=float).transpose()
            ll_intensity_sigma = numpy.array(ll_intensity_sigma, dtype=float).transpose()

        self.__dict__["gamma"] = l_gamma
        self.__dict__["nu"] = numpy.array(l_nu, dtype=float)
        self.__dict__["intensity_plus_net"] = ll_intensity_plus_net
        self.__dict__["intensity_minus_net"] = ll_intensity_minus_net
        self.__dict__["intensity_bkg_calc"] = ll_intensity_bkg_calc
        self.__dict__["excluded_points"] = ll_excluded_points

        if flag_polarized:
            self.__dict__["intensity_plus"] = ll_intensity_plus
            self.__dict__["intensity_plus_sigma"] = ll_intensity_plus_sigma
            self.__dict__["intensity_minus"] = ll_intensity_minus
            self.__dict__["intensity_minus_sigma"] = ll_intensity_minus_sigma
        elif flag_unpolarized:
            self.__dict__["intensity"] = ll_intensity
            self.__dict__["intensity_sigma"] = ll_intensity_sigma
        return flag

    def form_gamma_nu_intensity_plus_net(self) -> bool:
        if ((self.nu is not None) & (self.gamma is not None) & (self.intensity_plus_net is not None)):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.intensity_plus_net
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity_plus_net"] = "\n".join(ls_out)

    def form_gamma_nu_intensity_minus_net(self) -> bool:
        if ((self.nu is not None) & (self.gamma is not None) & (self.intensity_minus_net is not None)):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.intensity_minus_net
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity_minus_net"] = "\n".join(ls_out)

    def form_gamma_nu_excluded_points(self) -> bool:
        if ((self.nu is not None) & (self.gamma is not None) & (self.excluded_points is not None)):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.excluded_points
            ll_intensity = [["True" if ll_intensity[_2][_1] else "False" for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_excluded_points"] = "\n".join(ls_out)

    def form_gamma_nu_intensity_bkg_calc(self) -> bool:
        if ((self.nu is not None) & (self.gamma is not None) & (self.intensity_bkg_calc is not None)):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.intensity_bkg_calc
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity_bkg_calc"] = "\n".join(ls_out)

    def form_gamma_nu_intensity_plus(self) -> bool:
        if ((self.nu is not None) & (self.gamma is not None) & (self.intensity_plus is not None)):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.intensity_plus
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity_plus"] = "\n".join(ls_out)

    def form_gamma_nu_intensity_plus_sigma(self) -> bool:
        if ((self.nu is not None) & (self.gamma is not None) & (self.intensity_plus_sigma is not None)):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.intensity_plus_sigma
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity_plus_sigma"] = "\n".join(ls_out)

    def form_gamma_nu_intensity_minus(self) -> bool:
        if ((self.nu is not None) & (self.gamma is not None) & (self.intensity_minus is not None)):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.intensity_minus
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity_minus"] = "\n".join(ls_out)

    def form_gamma_nu_intensity_minus_sigma(self) -> bool:
        if ((self.nu is not None) & (self.gamma is not None) & (self.intensity_minus_sigma is not None)):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.intensity_minus_sigma
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity_minus_sigma"] = "\n".join(ls_out)

    def form_gamma_nu_intensity(self) -> bool:
        if ((self.nu is not None) & (self.gamma is not None) & (self.intensity is not None)):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.intensity
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity"] = "\n".join(ls_out)

    def form_gamma_nu_intensity_sigma(self) -> bool:
        if ((self.nu is not None) & (self.gamma is not None) & (self.intensity_sigma is not None)):
            ls_out = []
            ls_out.append("{:12} ".format(len(self.nu)) + " ".join(["{:6.2f}      ".format(_) for _ in self.gamma]))
            ll_intensity = self.intensity_sigma
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for nu, l_intensity in zip(self.nu, ll_intensity):
                ls_out.append("{:12.2f} ".format(nu) + " ".join(["{:12}".format(_) for _ in l_intensity]))
            self.__dict__["gamma_nu_intensity_sigma"] = "\n".join(ls_out)


    def plots(self):
        return [self.plot_gamma_nu()]
    
    def plot_gamma_nu(self):
        if not(self.is_attribute("gamma") & self.is_attribute("nu") &
               self.is_attribute("intensity_plus_net") & 
               self.is_attribute("intensity_minus_net")):
            return

        gamma = self.gamma 
        nu = self.nu 
        signal_plus = self.intensity_plus_net
        signal_minus = self.intensity_minus_net
        bkg_calc = self.intensity_bkg_calc
        signal_sum = signal_plus + signal_minus + bkg_calc
        signal_difference = signal_plus - signal_minus

        flag_polarized, flag_unpolarized = False, False
        if self.is_attribute("intensity_plus"):
            flag_polarized = True
        if self.is_attribute("intensity"):
            flag_unpolarized = True

        if flag_polarized:
            signal_exp_plus = self.intensity_plus
            signal_exp_minus = self.intensity_minus
            signal_exp_plus_sigma = self.intensity_plus_sigma
            signal_exp_minus_sigma = self.intensity_minus_sigma
            signal_exp_sum = signal_exp_plus + signal_exp_minus
            signal_exp_difference = signal_exp_plus - signal_exp_minus
            signal_exp_sum_sigma_sq = numpy.square(signal_exp_plus_sigma)+numpy.square(signal_exp_minus_sigma)
            signal_exp_sum_sigma = numpy.sqrt(signal_exp_sum_sigma_sq)
        elif flag_unpolarized:
            signal_exp_sum = self.intensity
            signal_exp_sum_sigma = self.intensity_sigma
            signal_exp_sum_sigma_sq = numpy.square(signal_exp_sum_sigma)

        excluded_points = numpy.logical_or(
            self.excluded_points,
            numpy.isnan(signal_exp_sum))
        included_points = numpy.logical_not(excluded_points)

        signal_em_sum = numpy.concatenate([numpy.flip(signal_sum, axis=1), signal_exp_sum], axis=1)
        if flag_polarized:
            signal_em_difference = numpy.concatenate([numpy.flip(signal_difference, axis=1), signal_exp_difference], axis=1)

        max_val = max([numpy.nanmax(signal_exp_sum[included_points]), numpy.nanmax(signal_sum[included_points])])
        min_val = numpy.nanmin(signal_em_sum)
        max_val = max_val - (max_val-min_val)*0.25
        n_sigma_threshold = 20.
        
        alpha_excl_points = numpy.zeros(signal_exp_sum.shape, dtype=int)
        alpha_excl_points[excluded_points]=200
        alpha_em_sum = numpy.concatenate([numpy.zeros_like(alpha_excl_points), alpha_excl_points], axis=1).transpose()

        zz_sum = numpy.stack([
            numpy.concatenate([numpy.zeros_like(alpha_excl_points), alpha_excl_points], axis=1).transpose(),
            numpy.concatenate([numpy.zeros_like(alpha_excl_points), alpha_excl_points], axis=1).transpose(),
            numpy.concatenate([numpy.zeros_like(alpha_excl_points), alpha_excl_points], axis=1).transpose(),
            alpha_em_sum],axis=2)

        extent = [gamma.min(), gamma.max(), nu.min(), nu.max()]
        cmap_sum = plt.get_cmap("turbo") # BuPu

        if flag_polarized:
            fig, axs = plt.subplots(2,2, sharex=True)
            ax1, ax2, ax3, ax4 = axs[0,0], axs[0,1], axs[1,0], axs[1,1]
        elif flag_unpolarized:
            fig, axs = plt.subplots(2, 1, sharex=True)
            ax1, ax3= axs[0], axs[1]

        norm_1 = matplotlib.colors.Normalize(vmax=max_val, vmin=min_val)
        img_unpol = ax1.imshow(signal_em_sum.transpose(), origin="lower", aspect="auto", cmap=cmap_sum, norm= norm_1, extent=extent)
        ax1.imshow(zz_sum, origin="lower", aspect="auto", alpha=0.8, extent=extent)

        ax1.set_xticks([])
        ax1.set_xlabel(r"$\gamma$ (deg.)")
        ax1.set_yticks([])
        ax1.set_ylabel("  Model       Experiment")

        if flag_polarized:
            cmap_difference = plt.get_cmap("turbo") # BrBG
            max_val = numpy.nanmax(numpy.abs(signal_em_difference))*0.50
            norm_2 = matplotlib.colors.Normalize(vmax=max_val, vmin=-max_val)
            ax2.imshow(signal_em_difference.transpose(), origin="lower", aspect="auto",
                      norm=norm_2,
                      cmap=cmap_difference,
                      extent=extent)

            ax2.set_xticks([])
            ax2.set_xlabel(r"$\gamma$ (deg.)")
            ax2.set_yticks([])

        ax1.set_ylabel("  Model       Experiment")

        chi_sq_per_n_sum = numpy.nansum(numpy.square(
                (signal_exp_sum-signal_sum)*included_points/signal_exp_sum_sigma))/numpy.sum(included_points)

        ax1.set_title(r"Unpolarized signal $\chi^2/n=$"+f"{chi_sq_per_n_sum:.2f}")


        ttheta, signal_projection_sum, signal_projection_exp_sum, signal_projection_exp_sum_sigma, \
            signal_projection_difference, signal_projection_exp_difference, signal_projection_exp_difference_sigma = \
            self.calc_projections_sum_difference()

        ax3.fill_between(
            ttheta[:gamma.size],(signal_projection_exp_sum-signal_projection_exp_sum_sigma)[:gamma.size],
            (signal_projection_exp_sum+signal_projection_exp_sum_sigma)[:gamma.size], color="k", alpha=0.4, label="experiment")

        # ax3.errorbar(
        #     ttheta[:gamma.size], signal_projection_exp_sum[:gamma.size],
        #     yerr=signal_projection_exp_sum_sigma[:gamma.size], fmt="ko", alpha=0.2, label="experiment")
        ax3.plot(ttheta[:gamma.size], signal_projection_sum[:gamma.size], "b-", label="model", linewidth=1)

        y_min_d, y_max_d = ax3.get_ylim()
        param = y_min_d-numpy.nanmax((signal_projection_exp_sum-signal_projection_sum)[:gamma.size])
        ax3.plot([ttheta[:gamma.size].min(), ttheta[:gamma.size].max()], [param, param], "k:")
        ax3.plot(ttheta[:gamma.size], (signal_projection_exp_sum-signal_projection_sum)[:gamma.size]+param,
            "r-", alpha=0.5) # , label="difference"

        # ax3.plot(ttheta[:gamma.size], signal_projection_exp_sum[:gamma.size], "b.", label="experiment")
        # ax3.plot(ttheta[:gamma.size], (signal_projection_exp_sum-signal_projection_sum)[:gamma.size], "r-")
        ax3.legend()
        ax3.set_xlabel(r"$2\theta$ (deg.)")
        # ax3.get_yaxis().set_visible(False)

        if flag_polarized:
            chi_sq_per_n_difference = numpy.nansum(numpy.square(
                    (signal_exp_difference-signal_difference)/signal_exp_sum_sigma))/numpy.product(signal_exp_difference.shape)
            ax2.set_title(r"Polarized signal $\chi^2/n=$"+f"{chi_sq_per_n_difference:.2f}")

            ax4.fill_between(
                ttheta[:gamma.size], (signal_projection_exp_difference-signal_projection_exp_difference_sigma)[:gamma.size],
                (signal_projection_exp_difference+signal_projection_exp_difference_sigma)[:gamma.size], color="k", alpha=0.4, label="experiment")
            # ax4.errorbar(
            #     ttheta[:gamma.size], signal_projection_exp_difference[:gamma.size],
            #     yerr=signal_projection_exp_difference_sigma[:gamma.size], fmt="ko", alpha=0.2, label="experiment")
            ax4.plot(ttheta[:gamma.size], signal_projection_difference[:gamma.size], "b-", label="model", linewidth=1)

            y_min_d, y_max_d = ax4.get_ylim()
            param = y_min_d-numpy.nanmax((signal_projection_exp_difference-signal_projection_difference)[:gamma.size])
            ax4.plot([ttheta[:gamma.size].min(), ttheta[:gamma.size].max()], [param, param], "k:")
            ax4.plot(ttheta[:gamma.size], (signal_projection_exp_difference-signal_projection_difference)[:gamma.size]+param,
                "r-", alpha=0.5) #, label="difference"
            ax4.legend()
            ax4.set_xlabel(r"$2\theta$ (deg.)")
            # ax4.get_yaxis().set_visible(False)
        
        return (fig, ax1)

    def calc_projections_sum_difference(self):
        gamma = self.gamma 
        nu = self.nu 

        gamma_rad, nu_rad = gamma*numpy.pi/180., nu*numpy.pi/180.
        ttheta_rad, phi_rad = calc_ttheta_phi_by_gamma_nu(
            gamma_rad[:, na], nu_rad[na, :], flag_gamma=False, flag_nu=False)[:2]
        ttheta_2d = ttheta_rad * 180./numpy.pi

        ttheta_step = numpy.round(gamma[1]-gamma[0], decimals=5)
        ttheta_min = gamma.min()
        ttheta_index = numpy.round((ttheta_2d-ttheta_min)/ttheta_step, decimals=0).astype(int)

        ttheta = numpy.linspace(
            ttheta_min,
            ttheta_step*ttheta_index.max()+ttheta_min,
            ttheta_index.max(), endpoint=False)

        signal_plus = self.intensity_plus_net
        signal_minus = self.intensity_minus_net
        bkg_calc = self.intensity_bkg_calc
        signal_sum = signal_plus + signal_minus + bkg_calc
        signal_difference = signal_plus - signal_minus

        flag_polarized, flag_unpolarized = False, False
        if self.is_attribute("intensity_plus"):
            flag_polarized = True
        elif self.is_attribute("intensity"):
            flag_unpolarized = True
        
        if flag_polarized:
            signal_exp_plus = self.intensity_plus
            signal_exp_minus = self.intensity_minus
            signal_exp_plus_sigma = self.intensity_plus_sigma
            signal_exp_minus_sigma = self.intensity_minus_sigma
            signal_exp_sum = signal_exp_plus + signal_exp_minus
            signal_exp_difference = signal_exp_plus - signal_exp_minus
            signal_exp_sum_sigma_sq = numpy.square(signal_exp_plus_sigma)+numpy.square(signal_exp_minus_sigma)
        elif flag_unpolarized:
            signal_exp_sum = self.intensity
            signal_exp_sum_sigma_sq = numpy.square(self.intensity_sigma)

        excluded_points = numpy.logical_or(
            self.excluded_points,
            numpy.isnan(signal_exp_sum))
        included_points = numpy.logical_not(excluded_points)

        signal_projection_sum = numpy.zeros_like(ttheta)
        signal_projection_exp_sum = numpy.zeros_like(ttheta)
        signal_projection_difference = numpy.zeros_like(ttheta)
        signal_projection_exp_sum_sigma = numpy.zeros_like(ttheta)
        signal_projection_exp_difference = numpy.zeros_like(ttheta)
        signal_projection_exp_difference_sigma = numpy.zeros_like(ttheta)

        if flag_polarized:
            flag_nonan = numpy.logical_not(numpy.isnan(signal_exp_difference))
        elif flag_polarized:
            flag_nonan = numpy.logical_not(numpy.isnan(signal_exp_sum))

        for ind in range(ttheta.size):
            flag_1 = ttheta_index == ind
            flag = numpy.logical_and(flag_1, included_points)
            points = numpy.sum(flag)
            if points != 0:
                signal_projection_sum[ind] += numpy.sum(signal_sum[flag])/points
                signal_projection_exp_sum[ind] += numpy.sum(signal_exp_sum[flag])/points
                signal_projection_exp_sum_sigma[ind] += numpy.sqrt(numpy.sum(signal_exp_sum_sigma_sq[flag]))/points
            else:
                signal_projection_sum[ind] = numpy.nan
                signal_projection_exp_sum[ind] = numpy.nan
                signal_projection_exp_sum_sigma[ind] = numpy.nan
            if flag_polarized:
                flag = numpy.logical_and(flag_1, flag_nonan)
                points = numpy.sum(flag)
                if points != 0:
                    signal_projection_difference[ind] += numpy.sum(signal_difference[flag])/points
                    signal_projection_exp_difference[ind] += numpy.sum(signal_exp_difference[flag])/points
                    signal_projection_exp_difference_sigma[ind] += numpy.sqrt(numpy.sum(signal_exp_sum_sigma_sq[flag]))/points
                else:
                    signal_projection_difference[ind] = numpy.nan
                    signal_projection_exp_difference[ind] = numpy.nan
                    signal_projection_exp_difference_sigma[ind] = numpy.nan
        
        
        return ttheta, signal_projection_sum, signal_projection_exp_sum, signal_projection_exp_sum_sigma, \
            signal_projection_difference, signal_projection_exp_difference, signal_projection_exp_difference_sigma

    def plot_diff_total(self):
        if not(self.is_attribute("gamma") & self.is_attribute("nu") &
               self.is_attribute("intensity_plus_total") & 
               self.is_attribute("intensity_minus_total")):
            return
        fig, ax = plt.subplots()
        ax.set_title("Polarized intensity")
        ax.set_xlabel("2 theta (degrees)")
        ax.set_ylabel('nu (degrees)')

        np_tth = numpy.array(self.gamma, dtype=float)
        np_nu = numpy.array(self.nu, dtype=float)
        np_up = self.intensity_plus_net
        np_down = self.intensity_minus_net
        np_diff = np_up - np_down
        
        norm = matplotlib.colors.Normalize(vmax=abs(np_diff).max(),
                                           vmin=-abs(np_diff).max())
        # cmap = matplotlib.PRGn
        
        extent =(np_tth.min(), np_tth.max(), np_nu.min(), np_nu.max())
        plt.set_cmap('rainbow')
        ax.imshow(np_diff.transpose(), interpolation='bilinear', extent=extent,
                  alpha=0.9, origin="lower", norm=norm)        
        
        # blk = '#000000'
        # ax.contour(np_tth, np_nu, np_sum.transpose(),
        #             #levels=[0.1, 0.5, 1., 5., 10., 50.],
        #             levels=5,
        #             colors=[blk, blk, blk, blk, blk, blk],
        #             linewidths=0.5)
        return (fig, ax)


    def estimate_gamma_nu_offsets(self):
        coeff_difference = 0.5
        gamma = self.gamma
        nu = self.nu
        flag_polarized, flag_unpolarized = False, False
        if self.is_attribute("intensity_plus"):
            flag_polarized = True
        elif self.is_attribute("intensity"):
            flag_unpolarized = True

        signal_sum = self.intensity_plus_net + self.intensity_minus_net + self.intensity_bkg_calc
        signal_difference = self.intensity_plus_net - self.intensity_minus_net
        
        if flag_polarized:
            signal_exp_sum = self.intensity_plus + self.intensity_minus 
            signal_exp_difference = self.intensity_plus - self.intensity_minus
            signal_exp_sum_sigma = numpy.sqrt(
                numpy.square(self.intensity_plus_sigma) +
                numpy.square(self.intensity_minus_sigma))
        elif flag_unpolarized:
            signal_exp_sum = self.intensity
            signal_exp_sum_sigma = self.intensity_sigma

        included_points = numpy.logical_not(self.excluded_points)

        def calc_chi_sq(param):
            offset_gamma = param[0]
            offset_nu = param[1]
            f_sum = interp2d(gamma, nu, signal_sum.transpose(), kind="linear")
            model_sum = f_sum(gamma-offset_gamma, nu-offset_nu).transpose()
            chi_sq_sum = numpy.nansum(numpy.square((model_sum-signal_exp_sum)/signal_exp_sum_sigma)[included_points])
            
            if flag_polarized:
                f_difference = interp2d(gamma, nu, signal_difference.transpose(), kind="linear")
                model_difference = f_difference(gamma-offset_gamma, nu-offset_nu).transpose()
                chi_sq_difference = numpy.nansum(numpy.square((model_difference-signal_exp_difference)/signal_exp_sum_sigma))
                chi_sq = (1. - coeff_difference) * chi_sq_sum + coeff_difference* chi_sq_difference
            elif flag_unpolarized:
                chi_sq = chi_sq_sum
            return chi_sq

        param = numpy.array([0., 0.], dtype=float)
        res = scipy.optimize.minimize(calc_chi_sq, param, method="Nelder-Mead")
        return res
    
    def estimate_background(self,
            pd2d_background: Pd2dBackground, offset_gamma: float = 0, offset_nu: float = 0):
        gamma = self.gamma
        nu = self.nu

        if self.is_attribute("intensity_plus"):
            signal_exp_sum = self.intensity_plus + self.intensity_minus 
            signal_exp_sum_sigma = numpy.sqrt(
                numpy.square(self.intensity_plus_sigma) +
                numpy.square(self.intensity_minus_sigma))
        elif self.is_attribute("intensity"):
            signal_exp_sum = self.intensity 
            signal_exp_sum_sigma = self.intensity_sigma

        signal_sum_net = self.intensity_plus_net + self.intensity_minus_net 
        included_points = numpy.logical_not(self.excluded_points)

        variable_names = pd2d_background.get_variable_names()
        param_0 = [pd2d_background.get_variable_by_name(name) for name in variable_names]
        def temp_func(param):
            for name, value in zip(variable_names, param):
                pd2d_background.set_variable_by_name(name, value)
            model_bkgr = pd2d_background.interpolate_by_points(
                gamma+offset_gamma, nu+offset_nu)
            chi_sq_sum = numpy.nansum(numpy.square((signal_exp_sum - signal_sum_net - model_bkgr)/signal_exp_sum_sigma)[included_points])
            return chi_sq_sum

        res = scipy.optimize.minimize(temp_func, param_0, method="Nelder-Mead")
        pd2d_background.form_gamma_nu_intensity()
        self.intensity_bkg_calc = pd2d_background.interpolate_by_points(gamma+offset_gamma, nu+offset_nu)
        self.form_gamma_nu_intensity_bkg_calc()
        return res


import numpy

from cryspy.B_parent_classes.cl_4_global import GlobalN

from cryspy.C_item_loop_classes.cl_1_chi2 import Chi2
from cryspy.C_item_loop_classes.cl_1_pd_background import PdBackgroundL
from cryspy.C_item_loop_classes.cl_1_pd_instr_reflex_asymmetry import PdInstrReflexAsymmetry
from cryspy.C_item_loop_classes.cl_1_pd_instr_resolution import PdInstrResolution
from cryspy.C_item_loop_classes.cl_1_pd_meas import PdMeasL
from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import DiffrnRadiation
from cryspy.C_item_loop_classes.cl_1_range import Range
from cryspy.C_item_loop_classes.cl_1_phase import PhaseL
from cryspy.C_item_loop_classes.cl_1_setup import Setup

from cryspy.E_data_classes.cl_1_crystal import Crystal
from cryspy.E_data_classes.cl_2_pd import Pd

from cryspy.procedure_rhochi.rhochi import rhochi_no_refinement

def simulation_pd(l_crystal:list,
                  ttheta_start:float=4.,
                  ttheta_end:float=120.,
                  ttheta_step:float=0.2,
                  flag_polarized:bool=False):
    l_crystal = [item for item in l_crystal if isinstance(item, Crystal)]
    n_step = int(numpy.round((ttheta_end-ttheta_start)/ttheta_step+1.))
    tth_deg = numpy.round(numpy.linspace(ttheta_start, ttheta_end, n_step), 3)
    bckg_pd = PdBackgroundL(
        intensity=[0.,0.],
        ttheta=[ttheta_start, ttheta_end])
    instr_res_pd = PdInstrResolution(w=0.1)
    phase_pd = PhaseL(scale=[1. for hh in l_crystal], label=[hh.data_name for hh in l_crystal])
    setup_pd = Setup(wavelength=1.4, offset_ttheta=0.)
    instr_asym_pd = PdInstrReflexAsymmetry(p1=0, p2=0, p3=0, p4=0)
    range_pd = Range(ttheta_min=ttheta_start, ttheta_max=ttheta_end)
    chi_sq = Chi2(sum=True, diff=False, up=False, down=False)
    l_tem_pd = [bckg_pd, instr_res_pd, phase_pd, setup_pd, instr_asym_pd, range_pd, chi_sq]
    if flag_polarized:
        meas_pd = PdMeasL(ttheta=tth_deg,
                                 intensity_plus=0*tth_deg,
                                 intensity_plus_sigma=0*tth_deg+0.1,
                                 intensity_minus=0*tth_deg,
                                 intensity_minus_sigma=0*tth_deg+0.1,)
        diffrn_rad = DiffrnRadiation(polarization=1., efficiency=1.)
        setup_pd.field = 1.
        l_tem_pd.extend([meas_pd, diffrn_rad])
    else:
        meas_pd = PdMeasL(ttheta=tth_deg, intensity=0*tth_deg, intensity_sigma=0*tth_deg+0.1)
        l_tem_pd.append(meas_pd)

    pd = Pd(data_name="sim_1d")
    pd.add_items(l_tem_pd)

    l_item_global = [pd, ] + l_crystal
    classes = set([type(item) for item in l_item_global])
    global_obj_2 = GlobalN.make_container((), tuple(classes), "global")
    global_obj_2.global_name = ""
    global_obj_2.add_items(l_item_global)

    rhochi_no_refinement(global_obj_2)

    pd_proc = pd.pd_proc
    pd_meas = pd.pd_meas
    if flag_polarized:
        pd_proc.numpy_intensity_plus = numpy.round(pd_proc.numpy_intensity_plus_net + 0.5 * pd_proc.numpy_intensity_bkg_calc, 3)
        pd_proc.numpy_intensity_minus = numpy.round(pd_proc.numpy_intensity_plus_net + 0.5 * pd_proc.numpy_intensity_bkg_calc, 3)
        pd_meas.numpy_intensity_plus = numpy.round(pd_proc.numpy_intensity_plus, 3)
        pd_meas.numpy_intensity_minus = numpy.round(pd_proc.numpy_intensity_minus, 3)
    else:
        pd_proc.numpy_intensity = numpy.round(pd_proc.numpy_intensity_plus_net + pd_proc.numpy_intensity_minus_net + pd_proc.numpy_intensity_bkg_calc, 3)
        pd_meas.numpy_intensity = numpy.round(pd_proc.numpy_intensity, 3)

    pd_proc.numpy_to_items()
    pd_meas.numpy_to_items()
    return pd


"""Description of classes Refln, ReflnL."""
from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class Refln(ItemN):
    """Nuclear structure factor of reflection.

    Data items in the REFLN category record details about the
    reflections used to determine the ATOM_SITE data items.
    The REFLN data items refer to individual reflections and must
    be included in looped lists.

    The REFLNS data items specify the parameters that apply to all
    reflections. The REFLNS data items are not looped.

    Mandatory attributes:
        - index_h, index_k, index_l

    Optional attributes:
        - a_calc, a_meas, b_calc, b_meas, class_code, crystal_id,
        - d_spacing, f_calc, f_meas, f_sigma, f_squared_calc,
        - f_squared_meas, f_squared_sigma, include_status,
        - intensity_calc, intensity_meas, intensity_sigma,
        - mean_path_length_tbar, phase_calc, phase_meas,
        - refinement_status, scale_group_code, sintlambda,
        - symmetry_epsilon, symmetry_multiplicity, wavelength,
        - wavelength_id

    """

    ATTR_MANDATORY_NAMES = ("index_h", "index_k", "index_l")
    ATTR_MANDATORY_TYPES = (int, int, int)
    ATTR_MANDATORY_CIF = ("index_h", "index_k", "index_l")

    ATTR_OPTIONAL_NAMES = (
        "a_calc", "a_meas", "b_calc", "b_meas", "class_code", "crystal_id",
        "d_spacing", "f_calc", "f_meas", "f_sigma", "f_squared_calc", "f_polarized_squared_calc",
        "f_squared_meas", "f_squared_sigma", "include_status",
        "intensity_calc", "intensity_meas", "intensity_sigma",
        "mean_path_length_tbar", "phase_calc", "phase_meas",
        "refinement_status", "scale_group_code", "sintlambda",
        "symmetry_epsilon", "symmetry_multiplicity", "wavelength",
        "wavelength_id")
    ATTR_OPTIONAL_TYPES = (float, float, float, float, str, str,
                           float, complex, complex, complex, float, float,
                           float, float, str, float, float, float,
                           float, str, str, str, str, float,
                           str, int, float, str)
    ATTR_OPTIONAL_CIF = (
        "a_calc", "a_meas", "b_calc", "b_meas", "class_code", "crystal_id",
        "d_spacing", "f_calc", "f_meas", "f_sigma", "f_squared_calc", "f_polarized_squared_calc",
        "f_squared_meas", "f_squared_sigma", "include_status",
        "intensity_calc", "intensity_meas", "intensity_sigma",
        "mean_path_length_tbar", "phase_calc", "phase_meas",
        "refinement_status", "scale_group_code", "sint/lambda",
        "symmetry_epsilon", "symmetry_multiplicity", "wavelength",
        "wavelength_id")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"f_calc": "{:.2f}", "a_calc": "{:.4f}", "b_calc": "{:.4f}",
                 "f_squared_calc": "{:.2f}", "f_polarized_squared_calc": "{:.2f}",
                 "sintlambda": "{:.5f}", "d_spacing": "{:.5f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {"include_status": ["o", "<", "-", "x", "h", "r"],
                     "refinement_status": ["incl", "excl", "extn"]}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "refln"

    def __init__(self, **kwargs) -> NoReturn:
        super(Refln, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"sintlambda": 0., "d_spacing": 0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class ReflnL(LoopN):
    """Nuclear structure factors for reflections."""

    ITEM_CLASS = Refln
    ATTR_INDEX = "id"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(ReflnL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name


# s_cont = """
#   loop_
#   _refln_index_h
#   _refln_index_k
#   _refln_index_l
#   _refln_d_spacing
#   _refln_A_calc
#   _refln_B_calc
#   _refln_f_calc
#   0 0 2 2.315 3.25  1.232 3.25+1.232j
#   2 2 0 4.213 5.00 -4.05 4.5
#   """

# obj = ReflnL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj.numpy_f_calc, end="\n\n")

# import numpy

# index_h = numpy.array([1,2,1,2], dtype=int)
# index_k = numpy.array([2,0,2,0], dtype=int)
# index_l = numpy.array([1,1,1,0], dtype=int)

# obj_2 = ReflnL()
# obj_2.numpy_index_h = index_h
# obj_2.numpy_index_k = index_k
# obj_2.numpy_index_l = index_l
# obj_2.numpy_to_items()
# print(obj_2)



"""DensityPoint and DensityPointL classes."""
from typing import NoReturn
from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

import numpy

from cryspy.A_functions_base.function_1_matrices import calc_mRmCmRT, \
    calc_phase_3d, calc_moment_2d_by_susceptibility
from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary
from cryspy.A_functions_base.function_1_strings import transform_string_to_r_b

from cryspy.A_functions_base.function_2_mem import \
    calc_asymmetric_unit_cell_indexes, calc_factor_in_front_of_density_for_fm,\
    calc_index_atom_symmetry_closest_to_fract_xyz, calc_moment_perp

from cryspy.C_item_loop_classes.cl_1_cell import Cell
from cryspy.C_item_loop_classes.cl_1_space_group_symop import SpaceGroupSymopL
from cryspy.C_item_loop_classes.cl_1_atom_site import AtomSiteL
from cryspy.C_item_loop_classes.cl_1_atom_site_susceptibility import \
    AtomSiteSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_atom_electron_configuration import \
    AtomElectronConfigurationL
from cryspy.C_item_loop_classes.cl_1_mem_parameters import MEMParameters
from cryspy.C_item_loop_classes.cl_2_space_group import SpaceGroup
from cryspy.C_item_loop_classes.cl_2_atom_rho_orbital_radial_slater import \
    AtomRhoOrbitalRadialSlaterL


class DensityPoint(ItemN):
    """
    Density point description.

    Attributes
    ----------
        - index_x, index_y, index_z (mandatory)
        - density, density_ferro, density_antiferro, multiplicity, fract_x,
          fract_y, fract_z, basin_atom_label, basin_atom_symop,
          basin_atom_distance (optional)

    Methods
    -------
        - calc_multiplicity
    """

    ATTR_MANDATORY_NAMES = ("index_x", "index_y", "index_z")
    ATTR_MANDATORY_TYPES = (int, int, int)
    ATTR_MANDATORY_CIF = ("index_x", "index_y", "index_z")

    ATTR_OPTIONAL_NAMES = ("density", "density_ferro", "density_antiferro",
                           "multiplicity", "fract_x", "fract_y", "fract_z",
                           "basin_atom_label", "basin_atom_symop",
                           "basin_atom_distance")
    ATTR_OPTIONAL_TYPES = (float, float, float, int, float, float, float,
                           str, str, float)
    ATTR_OPTIONAL_CIF = ("density", "density_ferro", "density_antiferro",
                         "multiplicity", "fract_x", "fract_y", "fract_z",
                         "basin_atom_label", "basin_atom_symop",
                         "basin_atom_distance")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"density": "{:.5f}", "density_ferro": "{:.5f}",
                 "density_antiferro": "{:.5f}", "fract_x": "{:.5f}",
                 "fract_y": "{:.5f}", "fract_z": "{:.5f}",
                 "basin_atom_distance": "{:.5f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "density_point"

    def __init__(self, **kwargs) -> NoReturn:
        super(DensityPoint, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"density": 0., "density_ferro": 0., "density_antiferro": 0.,
                 "multiplicity": 0, "basin_atom_distance": 0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_multiplicity(self, space_group: SpaceGroup):
        """Calculate the multiplicity of the point."""
        fract_x, fract_y, fract_z = self.fract_x, self.fract_y, self.fract_z
        mult = space_group.calc_xyz_mult(fract_x, fract_y, fract_z)[-1]
        # check of multiplicity
        el_2 = space_group.calc_el_symm_for_xyz(fract_x, fract_y,
                                                fract_z)[0].size
        mult_2 = int(len(space_group.full_space_group_symop.items)/el_2)
        if mult != mult_2:
            raise UserWarning(
                f"{fract_x:.5f} {fract_y:.5f} {fract_z:.5f} {mult:} {mult_2:}")
        self.multiplicity = mult


class DensityPointL(LoopN):
    """
    Description of Loop in loop.

    Internal Arguments
    ------------------
        - number_symmetry
        - number_unit_cell
        - volume_unit_cell

    Methods
    -------
        - form_asymmetric_unit_cell
    """

    ITEM_CLASS = DensityPoint
    ATTR_INDEX = None

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(DensityPointL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def form_asymmetric_unit_cell(self, space_group_symop: SpaceGroupSymopL,
                                  points_a: int = 48, points_b: int = 48,
                                  points_c: int = 48):
        """
        Form asymmetric unit cell.

        Parameters
        ----------
        space_group_symop : SpaceGroupSymop of cryspy library
            Full list of element symmetry.
        points_a : int, optional
            Number of points along axis a. The default is 48.
        points_b : int, optional
            Number of points along axis b. The default is 48.
        points_c : int, optional
            Number of points along axis c. The default is 48.

        Returns
        -------
        None.

        """
        r_11 = numpy.array(space_group_symop.r_11, dtype=float)
        r_12 = numpy.array(space_group_symop.r_12, dtype=float)
        r_13 = numpy.array(space_group_symop.r_13, dtype=float)

        r_21 = numpy.array(space_group_symop.r_21, dtype=float)
        r_22 = numpy.array(space_group_symop.r_22, dtype=float)
        r_23 = numpy.array(space_group_symop.r_23, dtype=float)

        r_31 = numpy.array(space_group_symop.r_31, dtype=float)
        r_32 = numpy.array(space_group_symop.r_32, dtype=float)
        r_33 = numpy.array(space_group_symop.r_33, dtype=float)

        b_1 = numpy.array(space_group_symop.b_1, dtype=float)
        b_2 = numpy.array(space_group_symop.b_2, dtype=float)
        b_3 = numpy.array(space_group_symop.b_3, dtype=float)

        r_ij = (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)
        b_i = (b_1, b_2, b_3)

        ind_x_a_u_c, ind_y_a_u_c, ind_z_a_u_c, counts_a_u_c = \
            calc_asymmetric_unit_cell_indexes(points_a, points_b, points_c,
                                              r_ij, b_i)

        fract_x_a_u_c = ind_x_a_u_c.astype(float)/float(points_a)
        fract_y_a_u_c = ind_y_a_u_c.astype(float)/float(points_b)
        fract_z_a_u_c = ind_z_a_u_c.astype(float)/float(points_c)

        l_item = []
        for _ix, _iy, _iz, _mult, _fx, _fy, _fz in zip(
                ind_x_a_u_c, ind_y_a_u_c, ind_z_a_u_c, counts_a_u_c,
                fract_x_a_u_c, fract_y_a_u_c, fract_z_a_u_c):

            item = DensityPoint(index_x=_ix, index_y=_iy, index_z=_iz,
                                multiplicity=_mult,
                                fract_x=_fx, fract_y=_fy, fract_z=_fz)
            l_item.append(item)
        self.items = l_item

        number_symmetry = r_11.size
        number_unit_cell = points_a*points_b*points_c

        self.number_symmetry = number_symmetry
        self.number_unit_cell = number_unit_cell

        self.calc_rbs_i(space_group_symop, points_a=points_a,
                        points_b=points_b, points_c=points_c)

    def separate_basins(self, space_group_symop: SpaceGroupSymopL, cell: Cell,
                        atom_site: AtomSiteL, points_a: int = 48,
                        points_b: int = 48, points_c: int = 48):
        """
        Separate unit cell on basins.

        Parameters
        ----------
        space_group_symop : SpaceGroupSymopL of cryspy library
            DESCRIPTION.
        cell : Cell of cryspy library
            DESCRIPTION.
        atom_site : AtomSiteL of cryspy library
            DESCRIPTION.
        points_a : int, optional
            DESCRIPTION. The default is 48.
        points_b : int, optional
            DESCRIPTION. The default is 48.
        points_c : int, optional
            DESCRIPTION. The default is 48.

        Returns
        -------
        None.

        """
        cell.form_object
        self.volume_unit_cell = cell.volume

        r_11 = numpy.array(space_group_symop.r_11, dtype=float)
        r_12 = numpy.array(space_group_symop.r_12, dtype=float)
        r_13 = numpy.array(space_group_symop.r_13, dtype=float)

        r_21 = numpy.array(space_group_symop.r_21, dtype=float)
        r_22 = numpy.array(space_group_symop.r_22, dtype=float)
        r_23 = numpy.array(space_group_symop.r_23, dtype=float)

        r_31 = numpy.array(space_group_symop.r_31, dtype=float)
        r_32 = numpy.array(space_group_symop.r_32, dtype=float)
        r_33 = numpy.array(space_group_symop.r_33, dtype=float)

        b_1 = numpy.array(space_group_symop.b_1, dtype=float)
        b_2 = numpy.array(space_group_symop.b_2, dtype=float)
        b_3 = numpy.array(space_group_symop.b_3, dtype=float)

        r_ij = (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)
        b_i = (b_1, b_2, b_3)

        fract_x = self.numpy_fract_x
        fract_y = self.numpy_fract_y
        fract_z = self.numpy_fract_z

        fract_xyz = numpy.transpose(numpy.vstack([fract_x, fract_y, fract_z]))

        fract_atom_x = numpy.array(atom_site.fract_x, dtype=float)
        fract_atom_y = numpy.array(atom_site.fract_y, dtype=float)
        fract_atom_z = numpy.array(atom_site.fract_z, dtype=float)

        fract_atom_xyz = numpy.transpose(numpy.vstack([
            fract_atom_x, fract_atom_y, fract_atom_z]))

        ind_at, ind_sym, numpy_basin_atom_distance = \
            calc_index_atom_symmetry_closest_to_fract_xyz(
                fract_xyz, fract_atom_xyz, r_ij, b_i, cell)

        numpy_atom_label = numpy.array(atom_site.label, str)
        numpy_basin_atom_label = numpy_atom_label[ind_at]

        numpy_symop_operation_xyz = numpy.array(
            space_group_symop.operation_xyz)
        numpy_basin_atom_symop = numpy_symop_operation_xyz[ind_sym]

        for item, basin_atom_label, basin_atom_symop, basin_atom_distance in \
            zip(self.items, numpy_basin_atom_label, numpy_basin_atom_symop,
                numpy_basin_atom_distance):
            item.basin_atom_label = str(basin_atom_label)
            item.basin_atom_symop = str(basin_atom_symop)
            item.basin_atom_distance = float(basin_atom_distance)

        numpy_structured_atom = numpy.array(
            list(zip(atom_site.label, atom_site.multiplicity)),
            dtype=[("label", "U10"), ("multiplicity", "i4")])

        self.numpy_structured_atom = numpy_structured_atom
        self.numpy_basin_atom_label = numpy_basin_atom_label
        self.numpy_basin_atom_symop = numpy_basin_atom_symop
        self.numpy_basin_atom_distance = numpy_basin_atom_distance

    def set_flat_density(self, l_magnetic_labes: list,
                         flag_two_channel: bool = False):
        r"""
        Define starting density as flat density by expression.

        p_{i} = \frac{N_{points} \Sum_{a} m_{a}} {V_{u.c.} \Sum_{i} m_{i}}
        """
        numpy_structured_atom = self.numpy_structured_atom
        n_u_c = self.number_unit_cell
        numpy_multiplicity = self.numpy_multiplicity

        volume_u_c = self.volume_unit_cell
        numpy_density = numpy.zeros(self.numpy_index_x.shape)
        numpy_basin_atom_label = self.numpy_basin_atom_label
        for structured_atom in numpy_structured_atom:
            mult_a = int(structured_atom["multiplicity"])
            label_a = str(structured_atom["label"])
            flag_atom = numpy_basin_atom_label == label_a
            sum_m_i_basin = (numpy_multiplicity[flag_atom]).sum()
            if sum_m_i_basin != 0:  # Case of two atoms in the same position
                den_i = float(n_u_c * mult_a) / float(sum_m_i_basin *
                                                      volume_u_c)
                numpy_density[flag_atom] = den_i

        self.numpy_density = numpy_density

        numpy_label = numpy.array(l_magnetic_labes, dtype=str)
        flag_2d = numpy_basin_atom_label[:, numpy.newaxis] == numpy_label[
            numpy.newaxis, :]
        flag_atom = numpy.any(flag_2d, axis=1)
        flag_background = numpy.logical_not(flag_atom)
        if flag_two_channel:
            flag_background = numpy.ones(flag_background.shape, bool)

        numpy_multiplicity = self.numpy_multiplicity

        sum_m_i_back = numpy_multiplicity[flag_background].sum()
        den_i = float(n_u_c * 1.) / float(sum_m_i_back * volume_u_c)
        numpy_density_ferro = numpy.zeros(self.numpy_index_x.shape)
        numpy_density_ferro[flag_background] += den_i

        numpy_density_antiferro = numpy.zeros(self.numpy_index_x.shape)
        numpy_density_antiferro[flag_background] += den_i

        self.numpy_density_ferro = numpy_density_ferro
        self.numpy_density_antiferro = numpy_density_antiferro

        self.renormalize_numpy_densities(flag_two_channel=flag_two_channel)
        self.numpy_to_items()

    def set_core_density(
            self, atom_site: AtomSiteL,
            atom_electron_configuration: AtomElectronConfigurationL,
            flag_two_channel: bool = False):
        r"""
        Define starting density as flat density by expression.

        p_{i} = \frac{N_{points} \Sum_{a} m_{a}} {V_{u.c.} \Sum_{i} m_{i}}
        """
        n_u_c = self.number_unit_cell
        volume_u_c = self.volume_unit_cell
        numpy_basin_atom_label = self.numpy_basin_atom_label
        numpy_basin_atom_distance = self.numpy_basin_atom_distance
        numpy_density = numpy.zeros(shape=numpy_basin_atom_distance.shape,
                                    dtype=float)

        for item_a_e_c in atom_electron_configuration.items:
            at_label = item_a_e_c.label
            at_type = atom_site[at_label].type_symbol

            flag_at_label = numpy_basin_atom_label == at_label

            l_sh_pop = item_a_e_c.get_core_shells_populations()

            l_arors = AtomRhoOrbitalRadialSlaterL().take_objects_for_atom_type(
                at_type)

            for sh_pop in l_sh_pop:
                shell, population = sh_pop
                obj = None
                for arors in l_arors:
                    try:
                        l_val = getattr(arors, f"coeff_{shell:}")
                        obj = arors
                        break
                    except AttributeError:
                        pass
                if obj is not None:
                    rho_sq = (obj.calc_normalized_rho(
                                numpy_basin_atom_distance[flag_at_label],
                                shell, kappa=1.))**2
                    rho_sq_1d = numpy.reshape(rho_sq, (rho_sq.size, ))

                    numpy_density[flag_at_label] += float(population) / \
                        (4.*numpy.pi) * rho_sq_1d

        self.numpy_density = numpy_density

        numpy_label = numpy.array([_h.label for _h in
                                   atom_electron_configuration.items],
                                  dtype=str)
        flag_2d = numpy_basin_atom_label[:, numpy.newaxis] == numpy_label[
            numpy.newaxis, :]
        flag_atom = numpy.any(flag_2d, axis=1)
        flag_background = numpy.logical_not(flag_atom)

        numpy_multiplicity = self.numpy_multiplicity
        sum_m_i_back = numpy_multiplicity[flag_background].sum()

        den_i = float(n_u_c * 1.) / float(sum_m_i_back * volume_u_c)
        numpy_density_ferro = numpy.zeros(numpy_basin_atom_distance.shape)
        numpy_density_ferro[flag_background] += den_i

        numpy_density_antiferro = numpy.zeros(numpy_basin_atom_distance.shape)
        numpy_density_antiferro[flag_background] += den_i

        self.numpy_density_ferro = numpy_density_ferro
        self.numpy_density_antiferro = numpy_density_antiferro

        self.renormalize_numpy_densities(flag_two_channel=flag_two_channel)
        self.numpy_to_items()


    def create_flat_density(self, space_group_symop: SpaceGroupSymopL,
                            cell: Cell, atom_site: AtomSiteL,
                            l_magnetic_labes: list, points_a: int = 48,
                            points_b: int = 48, points_c: int = 48,
                            flag_two_channel: bool = False):
        r"""Define starting density as flat density by expression.

        p_{i} = \frac{N_{points} \Sum_{a} m_{a}} {V_{u.c.} \Sum_{i} m_{i}}
        """
        self.form_asymmetric_unit_cell(
            space_group_symop, points_a=points_a, points_b=points_b,
            points_c=points_c)

        self.separate_basins(
            space_group_symop, cell, atom_site, points_a=points_a,
            points_b=points_b, points_c=points_c)

        self.set_flat_density(l_magnetic_labes,
                              flag_two_channel=flag_two_channel)

    def create_core_density(
            self, space_group_symop: SpaceGroupSymopL, cell: Cell,
            atom_site: AtomSiteL, atom_electron_configuration:
            AtomElectronConfigurationL, points_a: int = 48, points_b: int = 48,
            points_c: int = 48, flag_two_channel: bool = False):
        """Create core density.

        Parameters
        ----------
        space_group_symop : SpaceGroupSymopL
            DESCRIPTION.
        cell : Cell
            DESCRIPTION.
        atom_site : AtomSiteL
            DESCRIPTION.
        atom_electron_configuration : AtomElectronConfigurationL
            DESCRIPTION.
        l_magnetic_labes : TYPE
            DESCRIPTION.
        points_a : TYPE, optional
            DESCRIPTION. The default is 48.
        points_b : TYPE, optional
            DESCRIPTION. The default is 48.
        points_c : TYPE, optional
            DESCRIPTION. The default is 48.

        Returns
        -------
        None.

        """
        self.form_asymmetric_unit_cell(
            space_group_symop, points_a=points_a, points_b=points_b,
            points_c=points_c)

        self.separate_basins(
            space_group_symop, cell, atom_site, points_a=points_a,
            points_b=points_b, points_c=points_c)

        self.set_core_density(atom_site, atom_electron_configuration,
                              flag_two_channel=flag_two_channel)

    def calc_density_chi(self, cell: Cell,
                         atom_site_susceptibility: AtomSiteSusceptibilityL):
        """Calculate density chi."""
        density = self.density

        if atom_site_susceptibility is None:
            return numpy.zeros(len(density), dtype=float)

        label = self.basin_atom_label

        l_max_moment = [numpy.abs(
            item.calc_main_axes_of_magnetization_ellipsoid(cell)[0]).max()
            for item in atom_site_susceptibility.items]
        l_label_atom = atom_site_susceptibility.label

        l_den_chi = [den*l_max_moment[l_label_atom.index(lab)] if lab in
                     l_label_atom else 0. for lab, den in zip(label, density)]

        return numpy.array(l_den_chi, dtype=float)

    def save_to_file_den(
            self, mem_parameters: MEMParameters, space_group: SpaceGroup,
            cell: Cell, f_name: str = "file.den",
            atom_site_susceptibility: AtomSiteSusceptibilityL = None,
            f_background: str = "file_back.den"):
        """Save to file."""
        numpy_basin_atom_label = numpy.array(self.basin_atom_label, dtype=str)
        points_a = mem_parameters.points_a
        points_b = mem_parameters.points_b
        points_c = mem_parameters.points_c
        chi_iso_ferro = mem_parameters.chi_ferro
        chi_iso_antiferro = mem_parameters.chi_antiferro
        index_x = self.index_x
        index_y = self.index_y
        index_z = self.index_z
        # density = self.density
        density_chi = self.calc_density_chi(cell, atom_site_susceptibility)
        density_ferro = self.density_ferro
        density_antiferro = self.density_antiferro
        ls_out = []
        ls_out_b = ["Created by CrysPy, mu_B/Tesla"]
        ls_out.append("Created by CrysPy")
        ls_out.append("{:}".format(len(index_x)))
        ls_out_b.append("{:}".format(len(index_x)))

        for _x, _y, _z, _l, den_chi, den_f, den_a in \
            zip(index_x, index_y, index_z, numpy_basin_atom_label, density_chi,
                density_ferro, density_antiferro):
            ls_out.append(f"{_x:4} {_y:4} {_z:4} {den_chi:15.7f}")
            v1 = (chi_iso_ferro*den_f+chi_iso_antiferro*den_a)
            ls_out_b.append(f"{_x:4} {_y:4} {_z:4} {v1:15.7f}")

        a, b, c = cell.length_a, cell.length_b, cell.length_c
        al, be, ga = cell.angle_alpha, cell.angle_beta, cell.angle_gamma

        ls_out.append(
            f"{a:10.5f}{b:10.5f}{c:10.5f}{al:10.5f}{be:10.5f}{ga:10.5f}")
        ls_out_b.append(
            f"{a:10.5f}{b:10.5f}{c:10.5f}{al:10.5f}{be:10.5f}{ga:10.5f}")

        r_s_g_s = space_group.reduced_space_group_symop

        centr = int(space_group.centrosymmetry)
        l_orig = space_group.shift
        ls_out.append("{:5}{:5}{:5}{:5}{:5}{:5}".format(
            points_a, points_b, points_c, len(r_s_g_s.items), centr,
            len(l_orig)))
        ls_out_b.append("{:5}{:5}{:5}{:5}{:5}{:5}".format(
            points_a, points_b, points_c, len(r_s_g_s.items), centr,
            len(l_orig)))
        for _ in r_s_g_s.items:
            ls_out.append(
                f"{int(_.r_11):4}{int(_.r_21):4}{int(_.r_31):4}  \
{int(_.r_12):4}{int(_.r_22):4}{int(_.r_32):4}  {int(_.r_13):4}{int(_.r_23):4}\
{int(_.r_33):4}    {float(_.b_1):8.5f}{float(_.b_2):8.5f}{float(_.b_3):8.5f}")
            ls_out_b.append(
                f"{int(_.r_11):4}{int(_.r_21):4}{int(_.r_31):4}  \
{int(_.r_12):4}{int(_.r_22):4}{int(_.r_32):4}  {int(_.r_13):4}{int(_.r_23):4}\
{int(_.r_33):4}    {float(_.b_1):8.5f}{float(_.b_2):8.5f}{float(_.b_3):8.5f}")
        for orig in l_orig:
            ls_out.append(
                f"{float(orig[0]):8.4f}{float(orig[1]):8.4f}\
{float(orig[2]):8.4f}")
            ls_out_b.append(
                f"{float(orig[0]):8.4f}{float(orig[1]):8.4f}\
{float(orig[2]):8.4f}")

        with open(f_name, "w") as fid:
            fid.write("\n".join(ls_out))
        with open(f_background, "w") as fid:
            fid.write("\n".join(ls_out_b))
        return

    def calc_fm(self, factor_i, factor_ferro_i, factor_antiferro_i):
        """Calculate magnetic structure factor or its perpendicular component.

        Parameters
        ----------
        factor_i : TYPE [hkl, points]
            Factor is output of method calc_factor_in_front_of_density_for_fm
            for calculations of magnetic structure factor

            or

            it is output of method calc_factor_in_front_of_density_for_fm_perp
            for calculations of perpendicular component of magnetic structure
            factor.

        Returns
        -------
        msf_p_i : TYPE [hkl]
            Magnetic structure factor or perpendicular magnetic structure
            factor.
        delta_msf_i : TYPE [hkl, points]
            derivative over density

            d FM / d den

            or

            d FM_perp / d den.
        """
        np_den = self.numpy_density
        factor_x, factor_y, factor_z = factor_i

        f_m_perp_x = (factor_x * np_den[numpy.newaxis, :]).sum(axis=1)
        f_m_perp_y = (factor_y * np_den[numpy.newaxis, :]).sum(axis=1)
        f_m_perp_z = (factor_z * np_den[numpy.newaxis, :]).sum(axis=1)
        f_m_perp = (f_m_perp_x, f_m_perp_y, f_m_perp_z)

        delta_f_m_perp = (factor_x, factor_y, factor_z)

        np_den_f = self.numpy_density_ferro
        factor_f_x, factor_f_y, factor_f_z = factor_ferro_i

        f_m_perp_f_x = (factor_f_x * np_den_f[numpy.newaxis, :]).sum(axis=1)
        f_m_perp_f_y = (factor_f_y * np_den_f[numpy.newaxis, :]).sum(axis=1)
        f_m_perp_f_z = (factor_f_z * np_den_f[numpy.newaxis, :]).sum(axis=1)
        f_m_perp_f = (f_m_perp_f_x, f_m_perp_f_y, f_m_perp_f_z)

        delta_f_m_perp_f = (factor_f_x, factor_f_y, factor_f_z)

        np_den_a = self.numpy_density_antiferro

        factor_a_x, factor_a_y, factor_a_z = factor_antiferro_i

        f_m_perp_a_x = (factor_a_x * np_den_a[numpy.newaxis, :]).sum(axis=1)
        f_m_perp_a_y = (factor_a_y * np_den_a[numpy.newaxis, :]).sum(axis=1)
        f_m_perp_a_z = (factor_a_z * np_den_a[numpy.newaxis, :]).sum(axis=1)
        f_m_perp_a = (f_m_perp_a_x, f_m_perp_a_y, f_m_perp_a_z)

        delta_f_m_perp_a = (factor_a_x, factor_a_y, factor_a_z)

        f_m_p = (f_m_perp[0]+f_m_perp_f[0]+f_m_perp_a[0],
                 f_m_perp[1]+f_m_perp_f[1]+f_m_perp_a[1],
                 f_m_perp[2]+f_m_perp_f[2]+f_m_perp_a[2])
        return f_m_p, delta_f_m_perp, delta_f_m_perp_f, delta_f_m_perp_a

    def calc_phase_3d(self, hkl, space_group_symop: SpaceGroupSymopL):
        """Calculate phase 3d.

        Parameters
        ----------
        hkl : TYPE
            DESCRIPTION.
        space_group_symop : SpaceGroupSymopL
            DESCRIPTION.

        Returns
        -------
        phase_3d : TYPE
            DESCRIPTION.
        """
        fract_x = self.numpy_fract_x
        fract_y = self.numpy_fract_y
        fract_z = self.numpy_fract_z
        fract_xyz = numpy.transpose(numpy.vstack([fract_x, fract_y, fract_z]))

        r_11 = numpy.array(space_group_symop.r_11, dtype=float)
        r_12 = numpy.array(space_group_symop.r_12, dtype=float)
        r_13 = numpy.array(space_group_symop.r_13, dtype=float)

        r_21 = numpy.array(space_group_symop.r_21, dtype=float)
        r_22 = numpy.array(space_group_symop.r_22, dtype=float)
        r_23 = numpy.array(space_group_symop.r_23, dtype=float)

        r_31 = numpy.array(space_group_symop.r_31, dtype=float)
        r_32 = numpy.array(space_group_symop.r_32, dtype=float)
        r_33 = numpy.array(space_group_symop.r_33, dtype=float)

        b_1 = numpy.array(space_group_symop.b_1, dtype=float)
        b_2 = numpy.array(space_group_symop.b_2, dtype=float)
        b_3 = numpy.array(space_group_symop.b_3, dtype=float)

        r_ij = (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)
        b_i = (b_1, b_2, b_3)

        phase_3d = calc_phase_3d(hkl, r_ij, b_i, fract_xyz)
        return phase_3d

    def calc_rbs_i(self, space_group_symop: SpaceGroupSymopL,
                   points_a: int = 48, points_b: int = 48, points_c: int = 48):
        """
        Caclulate symmetry elements that transform point to itself.

        Parameters
        ----------
        space_group_symop : SpaceGroupSymopL
            DESCRIPTION.

        Returns
        -------
        None.

        """
        r_11 = numpy.array(space_group_symop.r_11, dtype=float)
        r_12 = numpy.array(space_group_symop.r_12, dtype=float)
        r_13 = numpy.array(space_group_symop.r_13, dtype=float)

        r_21 = numpy.array(space_group_symop.r_21, dtype=float)
        r_22 = numpy.array(space_group_symop.r_22, dtype=float)
        r_23 = numpy.array(space_group_symop.r_23, dtype=float)

        r_31 = numpy.array(space_group_symop.r_31, dtype=float)
        r_32 = numpy.array(space_group_symop.r_32, dtype=float)
        r_33 = numpy.array(space_group_symop.r_33, dtype=float)

        b_1 = numpy.array(space_group_symop.b_1, dtype=float)
        b_2 = numpy.array(space_group_symop.b_2, dtype=float)
        b_3 = numpy.array(space_group_symop.b_3, dtype=float)

        np_ind_x, np_ind_y = self.numpy_index_x, self.numpy_index_y
        np_ind_z = self.numpy_index_z
        if np_ind_x is None:
            np_ind_x = numpy.array(self.index_x, dtype=int)
            np_ind_y = numpy.array(self.index_y, dtype=int)
            np_ind_z = numpy.array(self.index_z, dtype=int)

        l_rs_i = []
        l_bs_i = []
        for i_x, i_y, i_z in zip(np_ind_x, np_ind_y, np_ind_z):
            np_x_n = numpy.mod((numpy.around((
                i_x*r_11 + i_y*r_12 + i_z*r_13 +
                points_a*b_1).astype(float), 0)).astype(int), points_a)
            np_y_n = numpy.mod((numpy.around((
                i_x*r_21 + i_y*r_22 + i_z*r_23 +
                points_b*b_2).astype(float), 0)).astype(int), points_b)
            np_z_n = numpy.mod((numpy.around((
                i_x*r_31 + i_y*r_32 + i_z*r_33 +
                points_c*b_3).astype(float), 0)).astype(int), points_c)
            flag_x = np_x_n == i_x
            flag_y = np_y_n == i_y
            flag_z = np_z_n == i_z
            flag_xy = numpy.logical_and(flag_x, flag_y)
            fl_xyz = numpy.logical_and(flag_xy, flag_z)
            r_11_c, r_12_c, r_13_c = r_11[fl_xyz], r_12[fl_xyz], r_13[fl_xyz]
            r_21_c, r_22_c, r_23_c = r_21[fl_xyz], r_22[fl_xyz], r_23[fl_xyz]
            r_31_c, r_32_c, r_33_c = r_31[fl_xyz], r_32[fl_xyz], r_33[fl_xyz]

            b_1_c, b_2_c, b_3_c = b_1[fl_xyz], b_2[fl_xyz], b_3[fl_xyz]
            l_rs_i.append((r_11_c, r_12_c, r_13_c, r_21_c, r_22_c, r_23_c,
                           r_31_c, r_32_c, r_33_c))
            l_bs_i.append((b_1_c, b_2_c, b_3_c))
        self.l_rs_i = l_rs_i
        self.l_bs_i = l_bs_i

    def calc_susc_i(self, atom_site_susceptibility: AtomSiteSusceptibilityL):
        """Calculate susceptibility of point i.

        Parameters
        ----------
        atom_site_susceptibility : AtomSiteSusceptibilityL
            DESCRIPTION.

        Returns
        -------
        susc_11 : TYPE
            DESCRIPTION.
        susc_22 : TYPE
            DESCRIPTION.
        susc_33 : TYPE
            DESCRIPTION.
        susc_12 : TYPE
            DESCRIPTION.
        susc_13 : TYPE
            DESCRIPTION.
        susc_23 : TYPE
            DESCRIPTION.
        """
        numpy_basin_atom_label = self.numpy_basin_atom_label
        numpy_basin_atom_symop = self.numpy_basin_atom_symop
        if numpy_basin_atom_label is None:
            numpy_basin_atom_label = numpy.array(self.basin_atom_label,
                                                 dtype=str)
            numpy_basin_atom_symop = numpy.array(self.basin_atom_symop,
                                                 dtype=str)
            self.numpy_basin_atom_label = numpy_basin_atom_label
            self.numpy_basin_atom_symop = numpy_basin_atom_symop

        np_label = numpy.array(atom_site_susceptibility.label, dtype=str)
        np_chi_11 = numpy.array(atom_site_susceptibility.chi_11, dtype=float)
        np_chi_22 = numpy.array(atom_site_susceptibility.chi_22, dtype=float)
        np_chi_33 = numpy.array(atom_site_susceptibility.chi_33, dtype=float)
        np_chi_12 = numpy.array(atom_site_susceptibility.chi_12, dtype=float)
        np_chi_13 = numpy.array(atom_site_susceptibility.chi_13, dtype=float)
        np_chi_23 = numpy.array(atom_site_susceptibility.chi_23, dtype=float)

        l_r = []
        for symop in numpy_basin_atom_symop:
            r, b = transform_string_to_r_b(symop)
            r = r.astype(float)
            l_r.append((r[0, 0], r[0, 1], r[0, 2], r[1, 0], r[1, 1], r[1, 2],
                        r[2, 0], r[2, 1], r[2, 2]))
        np_r = numpy.array(l_r, dtype=float)
        np_r_11, np_r_12, np_r_13 = np_r[:, 0], np_r[:, 1], np_r[:, 2]
        np_r_21, np_r_22, np_r_23 = np_r[:, 3], np_r[:, 4], np_r[:, 5]
        np_r_31, np_r_32, np_r_33 = np_r[:, 6], np_r[:, 7], np_r[:, 8]

        susc_11 = numpy.zeros(numpy_basin_atom_label.size, dtype=float)
        susc_22 = numpy.zeros(numpy_basin_atom_label.size, dtype=float)
        susc_33 = numpy.zeros(numpy_basin_atom_label.size, dtype=float)
        susc_12 = numpy.zeros(numpy_basin_atom_label.size, dtype=float)
        susc_13 = numpy.zeros(numpy_basin_atom_label.size, dtype=float)
        susc_23 = numpy.zeros(numpy_basin_atom_label.size, dtype=float)

        for _l, chi_11, chi_22, chi_33, chi_12, chi_13, chi_23 in \
                zip(np_label, np_chi_11, np_chi_22, np_chi_33, np_chi_12,
                    np_chi_13, np_chi_23):
            f_1 = numpy_basin_atom_label == _l
            chi_ij = (chi_11, chi_12, chi_13, chi_12, chi_22, chi_23,
                      chi_13, chi_23, chi_33)
            c_11, c_12, c_13, c_21, c_22, c_23, c_31, c_32, c_33 = \
                calc_mRmCmRT((np_r_11[f_1], np_r_12[f_1], np_r_13[f_1],
                              np_r_21[f_1], np_r_22[f_1], np_r_23[f_1],
                              np_r_31[f_1], np_r_32[f_1], np_r_33[f_1]),
                             chi_ij)

            susc_11[f_1] = c_11
            susc_22[f_1] = c_22
            susc_33[f_1] = c_33
            susc_12[f_1] = c_12
            susc_13[f_1] = c_13
            susc_23[f_1] = c_23
        l_rs_i = self.l_rs_i
        l_c_11, l_c_22, l_c_33, l_c_12, l_c_13, l_c_23 = [], [], [], [], [], []
        for s_11, s_22, s_33, s_12, s_13, s_23, rs_i in zip(
                susc_11, susc_22, susc_33, susc_12, susc_13, susc_23, l_rs_i):
            r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33 = rs_i
            c_11, c_12, c_13, c_21, c_22, c_23, c_31, c_32, c_33 = \
                calc_mRmCmRT(rs_i, (s_11, s_12, s_13,
                                    s_12, s_22, s_23,
                                    s_13, s_23, s_33))
            ns = c_11.size
            l_c_11.append(c_11.sum()/float(ns))
            l_c_22.append(c_22.sum()/float(ns))
            l_c_33.append(c_33.sum()/float(ns))
            l_c_12.append(c_12.sum()/float(ns))
            l_c_13.append(c_13.sum()/float(ns))
            l_c_23.append(c_23.sum()/float(ns))

        chi_11 = numpy.array(l_c_11, dtype=float)
        chi_22 = numpy.array(l_c_22, dtype=float)
        chi_33 = numpy.array(l_c_33, dtype=float)
        chi_12 = numpy.array(l_c_12, dtype=float)
        chi_13 = numpy.array(l_c_13, dtype=float)
        chi_23 = numpy.array(l_c_23, dtype=float)
        return chi_11, chi_22, chi_33, chi_12, chi_13, chi_23

    def calc_moment_2d(self, space_group_symop: SpaceGroupSymopL, cell: Cell,
                       atom_site_susceptibility: AtomSiteSusceptibilityL,
                       h_loc, chi_iso_ferro: float = 0.,
                       chi_iso_antiferro: float = 0.,
                       flag_two_channel: bool = False):
        """Calculate moments over points and symmetry elements.

        Parameters
        ----------
        space_group_symop : SpaceGroupSymopL
            DESCRIPTION.
        cell : Cell
            DESCRIPTION.
        atom_site_susceptibility : AtomSiteSusceptibilityL
            DESCRIPTION.
        h_loc : TYPE
            DESCRIPTION.
        chi_iso_ferro: TYPE
            Should be positive. Default is 0 mu_B per T per unit cell
        chi_iso_antiferro: TYPE
            Should be negative. Default is 0 mu_B per T per unit cell
        flag_back_all:bool
            True if background is defined for all points and
            False if background is defined for points where susceptibility is
            not defined

        Returns
        -------
        moment_2d : TYPE
            DESCRIPTION.
        moment_2d_ferro : TYPE
            DESCRIPTION.
        moment_2d_antiferro : TYPE
            DESCRIPTION.
        """
        cell.form_object
        m_norm = cell.m_m_norm
        m_norm_ij = (m_norm[0, 0], m_norm[0, 1], m_norm[0, 2],
                     m_norm[1, 0], m_norm[1, 1], m_norm[1, 2],
                     m_norm[2, 0], m_norm[2, 1], m_norm[2, 2])

        r_11 = numpy.array(space_group_symop.r_11, dtype=float)
        r_12 = numpy.array(space_group_symop.r_12, dtype=float)
        r_13 = numpy.array(space_group_symop.r_13, dtype=float)

        r_21 = numpy.array(space_group_symop.r_21, dtype=float)
        r_22 = numpy.array(space_group_symop.r_22, dtype=float)
        r_23 = numpy.array(space_group_symop.r_23, dtype=float)

        r_31 = numpy.array(space_group_symop.r_31, dtype=float)
        r_32 = numpy.array(space_group_symop.r_32, dtype=float)
        r_33 = numpy.array(space_group_symop.r_33, dtype=float)

        r_ij = (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)

        if not(flag_two_channel):
            s_11, s_22, s_33, s_12, s_13, s_23 = self.calc_susc_i(
                atom_site_susceptibility)

            susc_i = (0.2695*s_11, 0.2695*s_22, 0.2695*s_33,
                      0.2695*s_12, 0.2695*s_13, 0.2695*s_23)

            moment_2d = calc_moment_2d_by_susceptibility(
                r_ij, susc_i, m_norm_ij, h_loc)
        else:
            moment_2d = (
                numpy.zeros(shape=(self.numpy_index_x.size, r_ij[0].size)),
                numpy.zeros(shape=(self.numpy_index_x.size, r_ij[0].size)),
                numpy.zeros(shape=(self.numpy_index_x.size, r_ij[0].size)))

        moment_ferro = (0.2695*chi_iso_ferro*h_loc[0]*numpy.ones(
                            shape=(self.numpy_index_x.size, r_ij[0].size)),
                        0.2695*chi_iso_ferro*h_loc[1]*numpy.ones(
                            shape=(self.numpy_index_x.size, r_ij[0].size)),
                        0.2695*chi_iso_ferro*h_loc[2]*numpy.ones(
                            shape=(self.numpy_index_x.size, r_ij[0].size)))
        moment_antiferro = (0.2695*chi_iso_antiferro*h_loc[0]*numpy.ones(
                                shape=(self.numpy_index_x.size, r_ij[0].size)),
                            0.2695*chi_iso_antiferro*h_loc[1]*numpy.ones(
                                shape=(self.numpy_index_x.size, r_ij[0].size)),
                            0.2695*chi_iso_antiferro*h_loc[2]*numpy.ones(
                                shape=(self.numpy_index_x.size, r_ij[0].size)))

        return moment_2d, moment_ferro, moment_antiferro

    def calc_factor_in_front_of_density_for_fm(
            self, hkl, space_group_symop: SpaceGroupSymopL, cell: Cell,
            atom_site_susceptibility: AtomSiteSusceptibilityL, h_loc,
            chi_iso_ferro: float = 0., chi_iso_antiferro: float = 0.,
            flag_two_channel: bool = False):
        """Calculate factor in front of density for FM.

        Parameters
        ----------
        hkl : TYPE
            DESCRIPTION.
        space_group_symop : SpaceGroupSymopL
            DESCRIPTION.
        cell : Cell
            DESCRIPTION.
        atom_site_susceptibility : AtomSiteSusceptibilityL
            DESCRIPTION.
        h_loc : TYPE
            DESCRIPTION.
        chi_iso_ferro : float, optional
            DESCRIPTION. The default is 0..
        chi_iso_antiferro : float, optional
            DESCRIPTION. The default is 0..

        Returns
        -------
        v_hkl_2d_i : TYPE
            DESCRIPTION.
        v_ferro : TYPE
            DESCRIPTION.
        v_antiferro : TYPE
            DESCRIPTION.

        """
        mult_i = self.numpy_multiplicity
        np = self.number_unit_cell
        volume = self.volume_unit_cell
        moment_2d, moment_ferro, moment_antiferro = \
            self.calc_moment_2d(space_group_symop, cell,
                                atom_site_susceptibility, h_loc,
                                chi_iso_ferro=chi_iso_ferro,
                                chi_iso_antiferro=chi_iso_antiferro,
                                flag_two_channel=flag_two_channel)
        phase_3d = self.calc_phase_3d(hkl, space_group_symop)
        v_hkl_2d_i = \
            calc_factor_in_front_of_density_for_fm(mult_i, np, volume,
                                                   moment_2d, phase_3d, axis=2)
        v_ferro = \
            calc_factor_in_front_of_density_for_fm(
                mult_i, np, volume, moment_ferro, phase_3d, axis=2)

        v_antiferro = \
            calc_factor_in_front_of_density_for_fm(
                mult_i, np, volume, moment_antiferro, phase_3d, axis=2)

        return v_hkl_2d_i, v_ferro, v_antiferro

    def calc_factor_in_front_of_density_for_fm_perp(
            self, hkl, space_group_symop: SpaceGroupSymopL, cell: Cell,
            atom_site_susceptibility: AtomSiteSusceptibilityL, h_loc,
            chi_iso_ferro: float = 0., chi_iso_antiferro: float = 0.,
            flag_two_channel: bool = False):
        """Calculate factor in front of density for FM_perp.

        Arguments
        ---------
            - hkl
            - space_group_symop
            - ...
        """
        v_hkl_2d_i, v_ferro, v_antiferro = \
            self.calc_factor_in_front_of_density_for_fm(
                hkl, space_group_symop, cell, atom_site_susceptibility, h_loc,
                chi_iso_ferro=chi_iso_ferro,
                chi_iso_antiferro=chi_iso_antiferro,
                flag_two_channel=flag_two_channel)

        k_hkl_1, k_hkl_2, k_hkl_3 = cell.calc_k_loc(*hkl)
        na = numpy.newaxis
        v_hkl_perp_2d_i = calc_moment_perp(
            (k_hkl_1[:, na], k_hkl_2[:, na], k_hkl_3[:, na]), v_hkl_2d_i)
        v_perp_ferro = calc_moment_perp(
            (k_hkl_1[:, na], k_hkl_2[:, na], k_hkl_3[:, na]), v_ferro)
        v_perp_antiferro = calc_moment_perp(
            (k_hkl_1[:, na], k_hkl_2[:, na], k_hkl_3[:, na]), v_antiferro)
        return v_hkl_perp_2d_i, v_perp_ferro, v_perp_antiferro

    def calc_electrons_per_unit_cell(self):
        """
        Calculate electrons per unit cell.

        Returns
        -------
        numpy_atoms : TYPE
            DESCRIPTION.
        m_b_ferro : TYPE
            DESCRIPTION.
        m_b_aferro : TYPE
            DESCRIPTION.

        """
        numpy_density_ferro = self.numpy_density_ferro
        numpy_density_antiferro = self.numpy_density_antiferro
        numpy_multiplicity = self.numpy_multiplicity
        volume_unit_cell = self.volume_unit_cell
        number_unit_cell = self.number_unit_cell
        coeff = volume_unit_cell/float(number_unit_cell)
        m_b_ferro = (numpy_density_ferro*numpy_multiplicity).sum()*coeff
        m_b_aferro = (numpy_density_antiferro*numpy_multiplicity).sum()*coeff

        numpy_density = self.numpy_density
        numpy_basin_atom_label = self.numpy_basin_atom_label
        label_unique = numpy.unique(numpy_basin_atom_label)
        l_lab_m_at = []
        for atom_label in label_unique:
            f_at = (numpy_basin_atom_label == atom_label)
            m_at = (numpy_density[f_at]*numpy_multiplicity[f_at]).sum()*coeff
            l_lab_m_at.append((atom_label, m_at))

        numpy_atoms = numpy.array(l_lab_m_at, dtype=[("label", "U10"),
                                                     ("multiplicity", "f4")])

        return numpy_atoms, m_b_ferro, m_b_aferro

    def renormalize_numpy_densities(self, flag_two_channel: bool = False):
        """
        Renormalize numpy densitites.

        No input arguments. The internal arguments should be defiened.
        """
        numpy_density_ferro = self.numpy_density_ferro
        numpy_density_antiferro = self.numpy_density_antiferro
        numpy_multiplicity = self.numpy_multiplicity
        volume_unit_cell = self.volume_unit_cell
        number_unit_cell = self.number_unit_cell
        coeff = volume_unit_cell/float(number_unit_cell)
        m_b_ferro = (numpy_density_ferro*numpy_multiplicity).sum()*coeff
        m_b_aferro = (numpy_density_antiferro*numpy_multiplicity).sum()*coeff

        numpy_density_ferro /= m_b_ferro
        numpy_density_antiferro /= m_b_aferro

        if not(flag_two_channel):
            numpy_structured_atom = self.numpy_structured_atom

            numpy_density = self.numpy_density
            numpy_basin_atom_label = self.numpy_basin_atom_label

            for structured_atom in numpy_structured_atom:
                atom_label = structured_atom["label"]
                mult_a = structured_atom["multiplicity"]
                f_at = (numpy_basin_atom_label == atom_label)
                m_at = (numpy_density[f_at]*numpy_multiplicity[f_at]).sum() * \
                    coeff
                if m_at == 0.:
                    numpy_density[f_at] = float(mult_a)/(
                        coeff*float((numpy_multiplicity[f_at]).sum()))
                    m_at = float(mult_a)
                numpy_density[f_at] *= float(mult_a)/float(m_at)

# s_cont = """
#   loop_Ho1
#   _density_point_index_x
#   _density_point_index_y
#   _density_point_index_z
#   _density_point_density
#   _density_point_multiplicity
#   0  0  0  0.32  2
#   0  0  1  0.31  1
#   0  0  2  0.29  1
#   """

# obj = DensityPointL.from_cif(s_cont)
# print(obj, end="\n\n")



from typing import Callable
import numpy
import scipy
import scipy.optimize

from .rhochi_diffrn import calc_chi_sq_for_diffrn_by_dictionary
from .rhochi_pd import calc_chi_sq_for_pd_by_dictionary
from .rhochi_pd2d import calc_chi_sq_for_pd2d_by_dictionary
from .rhochi_tof import calc_chi_sq_for_tof_by_dictionary

from cryspy.A_functions_base.function_1_inversed_hessian import \
    estimate_inversed_hessian_matrix

na = numpy.newaxis



#     if optimization_method == "basinhopping":
#         # basinhopping
#         res = scipy.optimize.basinhopping(
#             tempfunc, param_0, niter=10, T=10, stepsize=0.1, interval=20,
#             disp=disp)
#         _dict_out = {"flag": flag, "res": res}
#
#     elif optimization_method == "simplex":
#         # simplex
#         res = scipy.optimize.minimize(
#             tempfunc, param_0, method='Nelder-Mead',
#             callback=lambda x: _f_callback(
#                 disp, coeff_norm, x, param_name=l_var_name, d_info=d_info),
#             options={"fatol": 0.01*n})
#         m_error, dist_hh = error_estimation_simplex(
#             res["final_simplex"][0], res["final_simplex"][1], tempfunc)
#         l_sigma = []
#         for i, val_2 in zip(range(m_error.shape[0]), dist_hh):
#             # slightly change definition, instead of (n-k) here is n
#             error = (abs(m_error[i, i])*1./n)**0.5
#             if m_error[i, i] < 0.:
#                 warn("Negative diagonal elements of Hessian.", UserWarning)
#             if val_2 > error:
#                 warn("Minimum is not found.", UserWarning)
#             l_sigma.append(max(error, val_2))
#         obj_hm = cryspy.InversedHessian()
#         obj_hm.set_inversed_hessian(m_error*1./float(n))
#         l_label = [var_name[-1][0] for var_name in l_var_name]
#         obj_hm.set_labels(l_label)
#         obj_hm.form_inversed_hessian()
#         obj_hm.form_object()
#         cryspy_object.inversed_hessian = obj_hm
#         for var_name, sigma, coeff in \
#                 zip(l_var_name, l_sigma, coeff_norm):
#             hh = tuple((f"{var_name[-1][0]:}_sigma", ) + var_name[-1][1:])
#             var_name_sigma = tuple(var_name[:-1]+(hh, ))
#             cryspy_object.set_variable_by_name(var_name_sigma, sigma*coeff)
#         _dict_out = {"flag": flag, "res": res}

def calc_punishement_function(dict_obj):
    chi_sq_punishement = 0.
    if "punishment_function" in dict_obj.keys():
        expression, d_marked = dict_obj["punishment_function"]
        d_value = {}
        for mark in d_marked.keys():
            way = d_marked[mark]
            d_value[mark] = dict_obj[way[0]][way[1]][way[2]]
        chi_sq_punishement += eval(expression, {}, d_value)
    return chi_sq_punishement


def calc_shift_p_by_constraints_hamilton(delta_p, dder_chi_sq, matrix_q):
    """Linear type Hamilton constraints.
    """
    q_t = matrix_q.transpose()
    q_dder_chi_sq = matrix_q.dot(dder_chi_sq)
    q_dder_chi_sq_qt = q_dder_chi_sq.dot(q_t)
    inv_q_dder_qt = numpy.linalg.inv(q_dder_chi_sq_qt)
    shift_delta_p = -delta_p.dot(q_t.dot(inv_q_dder_qt.dot(q_dder_chi_sq)))
    return shift_delta_p


def form_matrix_q(linear_constraints, parameter_name):
    matrix_q = numpy.zeros((len(linear_constraints), len(parameter_name)), dtype=float)
    for i_constraint, value_constraint in enumerate(linear_constraints):
        for coeff, p_name in value_constraint:
            ind_p_name = parameter_name.index(p_name)
            matrix_q[i_constraint, ind_p_name] = coeff
    return matrix_q


def rhochi_one_iteration_by_dictionary(
        global_dict, dict_in_out: dict = None, flag_use_precalculated_data: bool = False):
    chi_sq_sum, n_point_sum, der_chi_sq_sum, dder_chi_sq_sum, parameter_name_sum = \
        rhochi_calc_chi_sq_by_dictionary(
            global_dict, dict_in_out=dict_in_out, flag_use_precalculated_data=flag_use_precalculated_data,
            flag_calc_analytical_derivatives=True)

    delta_p = -1.* numpy.linalg.inv(dder_chi_sq_sum).dot(der_chi_sq_sum)
    if "linear_constraints" in global_dict.keys():
        linear_constraints = global_dict["linear_constraints"]
        matrix_q = form_matrix_q(linear_constraints, parameter_name_sum)
        shift_p = calc_shift_p_by_constraints_hamilton(delta_p, dder_chi_sq_sum, matrix_q)
        delta_p += shift_p
    for way, d_p in zip(parameter_name_sum, delta_p):
        global_dict[way[0]][way[1]][way[2]] += d_p
    return chi_sq_sum, n_point_sum, delta_p, parameter_name_sum, der_chi_sq_sum, dder_chi_sq_sum


def rhochi_rietveld_refinement_by_dictionary(global_dict: dict, method: str = "BFGS", callback: Callable = None):
    print("*********************************************")
    print("Rietveld refinement by CrysPy (module RhoChi)")
    print("*********************************************\n")
    print("Derivatives are calculated numerically.")
    dict_in_out = {}
    flag_use_precalculated_data = False
    flag_calc_analytical_derivatives = False
    print("Preliminary calculations...", end="\r")
    chi_sq, n_point, der_chi_sq, dder_chi_sq, parameter_names = rhochi_calc_chi_sq_by_dictionary(
        global_dict,
        dict_in_out=dict_in_out,
        flag_use_precalculated_data=flag_use_precalculated_data, flag_calc_analytical_derivatives=flag_calc_analytical_derivatives)


    if "linear_constraints" in global_dict.keys():
        linear_constraints = global_dict["linear_constraints"]
        # try:
        #     matrix_q = form_matrix_q(linear_constraints, parameter_names)
        #     lb = numpy.zeros((matrix_q.shape[0],), dtype=float)
        #     ub = numpy.zeros((matrix_q.shape[0],), dtype=float)
        #     linear_constraint = scipy.optimize.LinearConstraint(matrix_q, lb, ub)
        # except ValueError:
        #     print(f"User constraints are not used.")
    else:
        linear_constraints = []

    parameter_names_fixed = [linear_constraint[1][1] for linear_constraint in linear_constraints]
    parameter_names_free = [way for way  in parameter_names if not(way in parameter_names_fixed)]
    param_0 = [global_dict[way[0]][way[1]][way[2]] for way  in parameter_names_free]
    print(f"Started chi_sq per number of points is {chi_sq/n_point:.2f}.         ")
    if len(param_0) == 0:
        res = {}
        print(r"<b>UNSUCCESSFULL TRY<\b>")
        print("\nFor refinement procedure some parameters have to be set as refined.")
        print("Print parenthesis after parameter which heve to be refined.")
        print("Example: 1.23()")
        return chi_sq, parameter_names, dict_in_out, res
    print(f"Number of fitting parameters {len(param_0):}")
    for name, val in zip(parameter_names_free, param_0):
        print(f" - {name:}  {val:.5f}")

    if len(parameter_names_fixed) >0:
        print(f"Number of constrained parameters:")
    for name in parameter_names_fixed:
        print(f" - {name:}")


    flag_use_precalculated_data = True
    def tempfunc(l_param):
        for way, param in zip(parameter_names_free, l_param):
            global_dict[way[0]][way[1]][way[2]] = param
        # linear constraint on one parameter (temporary solution)
        for linear_constraint in linear_constraints:
            first_name = linear_constraint[0][1]
            second_name = linear_constraint[1][1]
            coeff = -linear_constraint[1][0]/linear_constraint[0][0]
            constr_param = coeff*l_param[parameter_names_free.index(first_name)]
            global_dict[second_name[0]][second_name[1]][second_name[2]] = constr_param

        chi_sq = rhochi_calc_chi_sq_by_dictionary(
            global_dict,
            dict_in_out=dict_in_out,
            flag_use_precalculated_data=flag_use_precalculated_data,
            flag_calc_analytical_derivatives=flag_calc_analytical_derivatives)[0]
        
        chi_sq_punishement = calc_punishement_function(global_dict)
        return chi_sq + chi_sq_punishement


    print("\nMinimization procedure of chi_sq is running... ", end="\r")
    #if linear_constraint is not None:
    #    method="SLSQP"
    #    print(f"User constraints are used in the refinement by method '{method:}'.")
    #    res = scipy.optimize.minimize(tempfunc, param_0, method=method, constraints=linear_constraint)
    #    hess_inv, np_first_der = estimate_inversed_hessian_matrix(tempfunc, res.x)
    #    if hess_inv is not None:
    #        res["hess_inv"] = hess_inv
    #else:
    res = scipy.optimize.minimize(tempfunc, param_0, method=method, callback=callback)
    print("Optimization is done.                          ", end="\n")

    print("Calculations for optimal parameters... ", end="\r")
    dict_in_out = {}
    flag_use_precalculated_data = False
    chi_sq, n_point = rhochi_calc_chi_sq_by_dictionary(
        global_dict,
        dict_in_out=dict_in_out,
        flag_use_precalculated_data=flag_use_precalculated_data, flag_calc_analytical_derivatives=flag_calc_analytical_derivatives)[:2]
    print(f"Optimal chi_sq per n is {chi_sq/n_point:.2f}", end="\n")

    return chi_sq, parameter_names, dict_in_out, res

def func_callback(*arg):
    print(arg)

def rhochi_lsq_by_dictionary(global_dict):
    print("*********************************************")
    print("Rietveld refinement by CrysPy (module RhoChi)")
    print("*********************************************\n")
    print("Derivatives are calculated analytically.")
    print("User constraints are working.")

    dict_in_out = {}
    dict_in_out_2 = {}
    flag_use_precalculated_data = False
    chi_sq_sum, n_point, delta_p, parameter_name_sum, der_chi_sq_sum, dder_chi_sq_sum = rhochi_one_iteration_by_dictionary(
        global_dict, dict_in_out=dict_in_out, flag_use_precalculated_data=flag_use_precalculated_data)
    flag = True
    flag_use_precalculated_data = True
    while flag:
        chi_sq_sum_2, n_point, delta_p_2, parameter_name_sum_2, der_chi_sq_sum_2, dder_chi_sq_sum_2 = rhochi_one_iteration_by_dictionary(
            global_dict, dict_in_out=dict_in_out_2, flag_use_precalculated_data=flag_use_precalculated_data)
        if chi_sq_sum_2 < chi_sq_sum:
            chi_sq_sum = chi_sq_sum_2
            delta_p = delta_p_2
            der_chi_sq_sum = der_chi_sq_sum_2
            dder_chi_sq_sum = dder_chi_sq_sum_2
            parameter_name_sum = parameter_name_sum_2
            dict_in_out = dict_in_out_2
        else:
            for way, d_p in zip(parameter_name_sum_2, delta_p_2):
                global_dict[way[0]][way[1]][way[2]] -= d_p
            for way, d_p in zip(parameter_name_sum, delta_p):
                global_dict[way[0]][way[1]][way[2]] -= d_p
            flag = False
    return chi_sq_sum, n_point, delta_p, parameter_name_sum, der_chi_sq_sum, dder_chi_sq_sum, dict_in_out


def rhochi_calc_chi_sq_by_dictionary(
        global_dict, dict_in_out:dict=None, flag_use_precalculated_data: bool=False,
        flag_calc_analytical_derivatives: bool = False):
    """Calculate chi_sq.
    """
    dict_in_out_keys = dict_in_out.keys()
    dict_keys = global_dict.keys()
    l_dict_crystal, l_dict_magcrystal = [], []
    l_dict_diffrn = []
    l_dict_pd, l_dict_pd2d, l_dict_tof = [], [], []
    for name_key in dict_keys:
        if name_key.startswith("crystal_"):
            l_dict_crystal.append((name_key, global_dict[name_key]))
        # if name_key.startswith("magcrystal_"):
        #     l_dict_magcrystal.append((name_key, global_dict[name_key]))
        if name_key.startswith("diffrn_"):
            l_dict_diffrn.append((name_key, global_dict[name_key]))
        if name_key.startswith("pd_"):
            l_dict_pd.append((name_key, global_dict[name_key]))
        if name_key.startswith("pd2d_"):
            l_dict_pd2d.append((name_key, global_dict[name_key]))
        if name_key.startswith("tof_"):
            l_dict_tof.append((name_key, global_dict[name_key]))

    l_chi_sq, l_der_chi_sq, l_dder_chi_sq = [], [], []
    l_n_point = []
    l_parameter_name, parameter_name_full = [], []
    dder = {}
    l_experiments, l_diff_chi = [], []
    for name_key_diffrn, dict_diffrn in l_dict_diffrn:
        if flag_use_precalculated_data and (name_key_diffrn in dict_in_out_keys):
            dict_in_out_diffrn = dict_in_out[name_key_diffrn]
        else:
            dict_in_out_diffrn = {}
            dict_in_out[name_key_diffrn] = dict_in_out_diffrn

        phase_label = dict_diffrn["phase_name"][0].lower()
        label_1 = f"crystal_{phase_label:}"
        label_2 = f"magcrystal_{phase_label:}"
        if label_1 in dict_keys:
            name_key_crystal = label_1
            dict_crystal = global_dict[label_1]
        elif label_2 in dict_keys:
            name_key_crystal = label_2
            dict_crystal = global_dict[label_2]
        elif (phase_label == "" and
                (len(l_dict_crystal)+len(l_dict_magcrystal)) > 0):
            name_key_crystal, dict_crystal = (l_dict_crystal+l_dict_magcrystal)[0]
        else:
            raise AttributeError(f"Phase {phase_label:} is not found")

        chi_sq, n_point, der_chi_sq, dder_chi_sq, parameter_name = calc_chi_sq_for_diffrn_by_dictionary(
            dict_diffrn, dict_crystal, dict_in_out=dict_in_out_diffrn,
            flag_use_precalculated_data=flag_use_precalculated_data,
            flag_calc_analytical_derivatives=flag_calc_analytical_derivatives)
        l_chi_sq.append(chi_sq)
        l_n_point.append(n_point)
        l_der_chi_sq.append(der_chi_sq)
        l_dder_chi_sq.append(dder_chi_sq)
        l_parameter_name.append(parameter_name)
        parameter_name_full.extend(parameter_name)
        l_experiments.append(name_key_diffrn)

    dict_crystals = [hh[1] for hh in l_dict_crystal]
    for name_key_exp, dict_exp in l_dict_pd + l_dict_pd2d + l_dict_tof:
        if flag_use_precalculated_data and (name_key_exp in dict_in_out_keys):
            dict_in_out_diffrn = dict_in_out[name_key_exp]
        else:
            dict_in_out_diffrn = {}
            dict_in_out[name_key_exp] = dict_in_out_diffrn

        if name_key_exp.startswith("pd_"):
            chi_sq, n_point, der_chi_sq, dder_chi_sq, parameter_name = \
                calc_chi_sq_for_pd_by_dictionary(
                    dict_exp, dict_crystals,
                    dict_in_out=dict_in_out_diffrn,
                    flag_use_precalculated_data=flag_use_precalculated_data,
                    flag_calc_analytical_derivatives=flag_calc_analytical_derivatives)
            l_chi_sq.append(chi_sq)
            l_n_point.append(n_point)
            l_der_chi_sq.append(der_chi_sq)
            l_dder_chi_sq.append(dder_chi_sq)
            l_parameter_name.append(parameter_name)
            parameter_name_full.extend(parameter_name)
            l_experiments.append(name_key_exp)

        elif name_key_exp.startswith("tof_"):
            chi_sq, n_point, der_chi_sq, dder_chi_sq, parameter_name = \
                calc_chi_sq_for_tof_by_dictionary(
                    dict_exp, dict_crystals,
                    dict_in_out=dict_in_out_diffrn,
                    flag_use_precalculated_data=flag_use_precalculated_data,
                    flag_calc_analytical_derivatives=flag_calc_analytical_derivatives)
            l_chi_sq.append(chi_sq)
            l_n_point.append(n_point)
            l_der_chi_sq.append(der_chi_sq)
            l_dder_chi_sq.append(dder_chi_sq)
            l_parameter_name.append(parameter_name)
            parameter_name_full.extend(parameter_name)
            l_experiments.append(name_key_exp)

        elif name_key_exp.startswith("pd2d_"):
            chi_sq, n_point, der_chi_sq, dder_chi_sq, parameter_name = \
                calc_chi_sq_for_pd2d_by_dictionary(
                    dict_exp, dict_crystals,
                    dict_in_out=dict_in_out_diffrn,
                    flag_use_precalculated_data=flag_use_precalculated_data,
                    flag_calc_analytical_derivatives=flag_calc_analytical_derivatives)

            l_chi_sq.append(chi_sq)
            l_n_point.append(n_point)
            l_der_chi_sq.append(der_chi_sq)
            l_dder_chi_sq.append(dder_chi_sq)
            l_parameter_name.append(parameter_name)
            parameter_name_full.extend(parameter_name)
            l_experiments.append(name_key_exp)


    chi_sq_sum = sum(l_chi_sq) #Unity weighting scheme
    n_point_sum = sum(l_n_point)

    parameter_name_sum = list(set(parameter_name_full))
    der_chi_sq_sum = numpy.zeros((len(parameter_name_sum),), dtype=float)
    dder_chi_sq_sum = numpy.zeros((len(parameter_name_sum), len(parameter_name_sum)), dtype=float)
    if flag_calc_analytical_derivatives:
        for l_pn, dc, ddc in zip(l_parameter_name, l_der_chi_sq, l_dder_chi_sq):
            for i_pn_1, pn_1 in enumerate(l_pn):
                ind_1 = parameter_name_sum.index(pn_1)
                der_chi_sq_sum[ind_1] += dc[i_pn_1]
                dder_chi_sq_sum[ind_1, ind_1] += ddc[i_pn_1, i_pn_1]
                if len(l_pn) > (i_pn_1+1):
                    for i_pn_2, pn_2 in enumerate(l_pn[i_pn_1+1:]):
                        ind_2 = parameter_name_sum.index(pn_2)
                        dder_chi_sq_sum[ind_1, ind_2] += ddc[i_pn_1, i_pn_2+i_pn_1+1]
                        dder_chi_sq_sum[ind_2, ind_1] += ddc[i_pn_2+i_pn_1+1, i_pn_1]
    return chi_sq_sum, n_point_sum, der_chi_sq_sum, dder_chi_sq_sum, parameter_name_sum



"""Calculation for MagCrystal.

Functions
---------
    - calc_b_iso_beta
    - calc_f_nucl
    - calc_f_mag
"""
import numpy

from cryspy.A_functions_base.function_3_mcif import \
    calc_full_sym_elems, calc_multiplicity, calc_moment_by_sym_elem

from cryspy.A_functions_base.function_2_crystallography_base import \
    calc_phase_by_hkl_xyz_rb, calc_dwf

from cryspy.A_functions_base.symmetry_elements import calc_full_mag_elems

from cryspy.C_item_loop_classes.cl_1_cell import Cell
from cryspy.C_item_loop_classes.cl_1_atom_site import AtomSiteL
from cryspy.C_item_loop_classes.cl_1_atom_site_aniso import \
    AtomSiteAnisoL

from cryspy.C_item_loop_classes.cl_1_space_group_symop_magn_centering import \
    SpaceGroupSymopMagnCenteringL

from cryspy.C_item_loop_classes.cl_2_space_group_symop_magn_operation import \
    SpaceGroupSymopMagnOperationL

from cryspy.C_item_loop_classes.cl_1_atom_site_moment import AtomSiteMomentL

from cryspy.C_item_loop_classes.cl_2_atom_site_scat import AtomSiteScatL


def calc_b_iso_beta(cell: Cell, atom_site: AtomSiteL,
                    atom_site_aniso: AtomSiteAnisoL):
    """Calculate b_iso and beta_ij based on atom_site and atom_sites.

    For each atom defined in atom_site.
    """
    a_s = atom_site
    a_s_a = atom_site_aniso
    l_b_iso, l_beta = [], []
    coeff = float(8.*numpy.pi**2)
    for item_a_s in a_s.items:
        label_atom = item_a_s.label
        try:
            adp_type = item_a_s.adp_type
        except AttributeError:
            adp_type = None
        b_iso = 0.
        beta = (0., 0., 0., 0., 0., 0.)
        if adp_type == "Uiso":
            u_iso = float(item_a_s.u_iso_or_equiv)
            b_iso = float(8.*numpy.pi**2*u_iso)
        elif adp_type == "Biso":
            b_iso = float(item_a_s.b_iso_or_equiv)
        elif adp_type == "Uovl":
            # FIXME: correct it
            u_iso = float(item_a_s.u_iso_or_equiv)
            b_iso = coeff*u_iso
        elif adp_type == "Umpe":
            # FIXME: correct it
            u_iso = float(item_a_s.u_iso_or_equiv)
            b_iso = float(8.*numpy.pi**2*u_iso)
        elif adp_type == "Uani":
            item_a_s_a = a_s_a[label_atom]
            beta = item_a_s_a.calc_beta(cell)
        elif adp_type == "Bovl":
            # FIXME: correct it
            b_iso = float(item_a_s.b_iso_or_equiv)
        elif adp_type == "Bani":
            item_a_s_a = a_s_a[label_atom]
            beta = (float(item_a_s_a.b_11), float(item_a_s_a.b_22),
                    float(item_a_s_a.b_33), float(item_a_s_a.b_12),
                    float(item_a_s_a.b_13), float(item_a_s_a.b_23))
        l_b_iso.append(b_iso)
        l_beta.append(beta)
    np_b_iso = numpy.array(l_b_iso, dtype=float)
    np_beta = numpy.array(l_beta, dtype=float)
    return np_b_iso, np_beta


# FIXME: full_space_group_symop is temporary slow solution.
def calc_f_nucl(
        index_hkl,
        space_group_symop_magn_operation: SpaceGroupSymopMagnOperationL,
        space_group_symop_magn_centering: SpaceGroupSymopMagnCenteringL,
        cell: Cell, atom_site: AtomSiteL, atom_site_aniso: AtomSiteAnisoL,
        flag_derivatives: bool = False):
    """
    Calculate nuclear structure factor. TEST.

    Keyword Arguments
    -----------------
        index_h, index_k, index_l: 1D numpy array of Miller indexes

    Output
    ------
        f_nucl: 1D numpy array of Nuclear structure factor

    Example
    -------
        >>> import numpy as np
        >>> h, k, l = np.array([1,2],dtype=int), np.array([1,0],dtype=int),
                      np.array([1,0],dtype=int)
        >>> f_nucl, der = crystal.calc_f_nucl(h, k, l)

    """
    dder = {}

    se_sgsmo = space_group_symop_magn_operation.get_sym_elems()
    se_sgsmc = space_group_symop_magn_centering.get_sym_elems()

    elem_symm_fs = calc_full_mag_elems(se_sgsmo, se_sgsmc)


    occupancy = numpy.array(atom_site.occupancy, dtype=float)
    x = numpy.array(atom_site.fract_x, dtype=float)
    y = numpy.array(atom_site.fract_y, dtype=float)
    z = numpy.array(atom_site.fract_z, dtype=float)

    fract_xyz = numpy.array([x, y, z], dtype=float)
    # FIXME: temporary solution
    try:
        atom_multiplicity = numpy.array(atom_site.multiplicity, dtype=int)
    except AttributeError:
        atom_multiplicity = calc_multiplicity(elem_symm_fs, fract_xyz)

    scat_length_neutron = numpy.array(atom_site.scat_length_neutron,
                                      dtype=complex)

    occ_mult = occupancy*atom_multiplicity

    r_11 = elem_symm_fs[4].astype(float)
    r_12 = elem_symm_fs[5].astype(float)
    r_13 = elem_symm_fs[6].astype(float)
    r_21 = elem_symm_fs[7].astype(float)
    r_22 = elem_symm_fs[8].astype(float)
    r_23 = elem_symm_fs[9].astype(float)
    r_31 = elem_symm_fs[10].astype(float)
    r_32 = elem_symm_fs[11].astype(float)
    r_33 = elem_symm_fs[12].astype(float)
    b_1 = elem_symm_fs[0].astype(float)/elem_symm_fs[3].astype(float)
    b_2 = elem_symm_fs[1].astype(float)/elem_symm_fs[3].astype(float)
    b_3 = elem_symm_fs[2].astype(float)/elem_symm_fs[3].astype(float)

    phase_3d = calc_phase_by_hkl_xyz_rb(
        index_hkl, x, y, z, r_11, r_12, r_13, r_21, r_22,
        r_23, r_31, r_32, r_33, b_1, b_2, b_3)

    b_iso, beta = calc_b_iso_beta(cell, atom_site, atom_site_aniso)

    dwf_3d = calc_dwf(cell, index_hkl, b_iso, beta, r_11,
                      r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)

    hh = phase_3d*dwf_3d
    phase_2d = hh.sum(axis=2)  # sum over symmetry

    b_scat_2d = numpy.meshgrid(index_hkl[0], scat_length_neutron,
                               indexing="ij")[1]
    occ_mult_2d = numpy.meshgrid(index_hkl[0], occ_mult, indexing="ij")[1]

    hh = phase_2d * b_scat_2d * occ_mult_2d
    # nuclear structure factor in assymetric unit cell
    f_hkl_as = hh.sum(axis=1)*1./r_11.size

    # f_nucl = space_group.calc_f_hkl_by_f_hkl_as(index_h, index_k, index_l,
    #                                             f_hkl_as)

    return f_hkl_as, dder


def calc_f_mag(
        index_hkl,
        space_group_symop_magn_operation: SpaceGroupSymopMagnOperationL,
        space_group_symop_magn_centering: SpaceGroupSymopMagnCenteringL,
        cell: Cell, atom_site: AtomSiteL, atom_site_aniso: AtomSiteAnisoL,
        atom_site_scat: AtomSiteScatL, atom_site_moment: AtomSiteMomentL,
        flag_derivatives: bool = False, flag_only_orbital: bool = False):
    """Calculate magnetic structure factor.

    It's given in Cartesian coordianate system x||a* z||c
    (!!!!!!it should be checked!!!!!)

    Unity is 10**-12 cm
    """
    dder = {}
    index_h, index_k, index_l = index_hkl[0], index_hkl[1], index_hkl[2]

    m_x = numpy.array(atom_site_moment.crystalaxis_x, dtype=float)
    m_y = numpy.array(atom_site_moment.crystalaxis_y, dtype=float)
    m_z = numpy.array(atom_site_moment.crystalaxis_z, dtype=float)
    moment = numpy.array([m_x, m_y, m_z], dtype=float)

    l_it_a_s_mag = [atom_site[item.label] for item in atom_site_moment.items]
    atom_site_mag = AtomSiteL()
    atom_site_mag.items = l_it_a_s_mag

    l_x, l_y, l_z, l_occ, l_mult = [], [], [], [], []
    flag_calc_multiplicity = False
    for atom_site_item in atom_site_mag.items:
        l_x.append(atom_site_item.fract_x)
        l_y.append(atom_site_item.fract_y)
        l_z.append(atom_site_item.fract_z)
        l_occ.append(atom_site_item.occupancy)
        try:
            l_mult.append(atom_site_item.multiplicity)
        except AttributeError:
            flag_calc_multiplicity = True

    occupancy = numpy.array(l_occ, dtype=float)
    x = numpy.array(l_x, dtype=float)
    y = numpy.array(l_y, dtype=float)
    z = numpy.array(l_z, dtype=float)

    fract_xyz = numpy.array([x, y, z], dtype=float)

    b_iso, beta = calc_b_iso_beta(cell, atom_site_mag, atom_site_aniso)

    sthovl = cell.calc_sthovl(index_h, index_k, index_l)
    form_factor = atom_site_scat.calc_form_factor(
        sthovl, flag_only_orbital=flag_only_orbital)

    se_sgsmo = space_group_symop_magn_operation.get_sym_elems()
    se_sgsmc = space_group_symop_magn_centering.get_sym_elems()

    elem_symm_fs = calc_full_mag_elems(se_sgsmo, se_sgsmc)

    if flag_calc_multiplicity:
        atom_multiplicity = calc_multiplicity(elem_symm_fs[:13], fract_xyz)
    else:
        atom_multiplicity = numpy.array(l_mult, dtype=int)
    

    occ_mult = occupancy*atom_multiplicity

    r_11 = elem_symm_fs[4].astype(float)
    r_12 = elem_symm_fs[5].astype(float)
    r_13 = elem_symm_fs[6].astype(float)
    r_21 = elem_symm_fs[7].astype(float)
    r_22 = elem_symm_fs[8].astype(float)
    r_23 = elem_symm_fs[9].astype(float)
    r_31 = elem_symm_fs[10].astype(float)
    r_32 = elem_symm_fs[11].astype(float)
    r_33 = elem_symm_fs[12].astype(float)
    b_1 = elem_symm_fs[0].astype(float)/elem_symm_fs[3].astype(float)
    b_2 = elem_symm_fs[1].astype(float)/elem_symm_fs[3].astype(float)
    b_3 = elem_symm_fs[2].astype(float)/elem_symm_fs[3].astype(float)

    # [xyz, symm, mag_at]
    moments_3d = calc_moment_by_sym_elem(elem_symm_fs, moment)
    # print("moments_3d: \n", moments_3d)
    # print(get_str_for_sym_elem(sym_elems), end=2*"\n")

    phase_3d = calc_phase_by_hkl_xyz_rb(
        index_hkl, x, y, z, r_11, r_12, r_13, r_21, r_22,
        r_23, r_31, r_32, r_33, b_1, b_2, b_3)

    dwf_3d = calc_dwf(cell, index_hkl, b_iso, beta, r_11,
                      r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)

    n_a = numpy.newaxis

    hh_1 = phase_3d*dwf_3d*(moments_3d[0, :, :]).transpose()[n_a, :, :]
    hh_2 = phase_3d*dwf_3d*(moments_3d[1, :, :]).transpose()[n_a, :, :]
    hh_3 = phase_3d*dwf_3d*(moments_3d[2, :, :]).transpose()[n_a, :, :]

    phase_2d_1 = hh_1.sum(axis=2)  # sum over symmetry
    phase_2d_2 = hh_2.sum(axis=2)  # sum over symmetry
    phase_2d_3 = hh_3.sum(axis=2)  # sum over symmetry

    occ_mult_2d = numpy.meshgrid(index_h, occ_mult, indexing="ij")[1]

    # print("phase_2d_1:", phase_2d_1)

    hh_1 = phase_2d_1 * form_factor * occ_mult_2d
    hh_2 = phase_2d_2 * form_factor * occ_mult_2d
    hh_3 = phase_2d_3 * form_factor * occ_mult_2d
    # nuclear structure factor in assymetric unit cell
    f_hkl_1 = hh_1.sum(axis=1)*1./r_11.size
    f_hkl_2 = hh_2.sum(axis=1)*1./r_11.size
    f_hkl_3 = hh_3.sum(axis=1)*1./r_11.size

    # f_nucl = space_group.calc_f_hkl_by_f_hkl_as(index_h, index_k, index_l,
    #                                             f_hkl_as)
    # f_mag should be given in Cartezian coordinate system (x||a*, z||c)
    m_m = cell.m_m_norm
    f_hkl_cart_1 = (m_m[0, 0]*f_hkl_1 + m_m[0, 1]*f_hkl_2 + m_m[0, 2]*f_hkl_3
                    )*0.2695
    f_hkl_cart_2 = (m_m[1, 0]*f_hkl_1 + m_m[1, 1]*f_hkl_2 + m_m[1, 2]*f_hkl_3
                    )*0.2695
    f_hkl_cart_3 = (m_m[2, 0]*f_hkl_1 + m_m[2, 1]*f_hkl_2 + m_m[2, 2]*f_hkl_3
                    )*0.2695
    f_mag = numpy.array([f_hkl_cart_1, f_hkl_cart_2, f_hkl_cart_3],
                        dtype=complex)
    return f_mag, dder



import numpy
import matplotlib.pyplot as plt
from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class TOFIntensityIncident(ItemN):
    """Correction of incident intensity for time-of-flight experiment.

    Attributes
    ----------
        - a1, a2, a3, a4, a5, a6, a7, a8 (mandatory)
        - spectrum (optional)
        

    spectrum is "Maxwell" (default), "Empirical-Exponents"
        """
    ATTR_MANDATORY_NAMES = ("a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7",
                            "a8")
    ATTR_MANDATORY_TYPES = (float, float, float, float, float, float, float,
                            float, float)
    ATTR_MANDATORY_CIF = ("a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7", "a8")

    ATTR_OPTIONAL_NAMES = ("spectrum", )
    ATTR_OPTIONAL_TYPES = (str, )
    ATTR_OPTIONAL_CIF = ("spectrum", )

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7", "a8")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {}

    # constraints on the parameters
    D_CONSTRAINTS = {"spectrum": ["Maxwell", "Empirical-Exponents"]}

    # default values for the parameters
    D_DEFAULT = {"spectrum": "Maxwell", "a0": 0., "a1": 2339., "a2": 35.59e6,
                 "a3": 1., "a4": 0.2015e-6, "a5": 0.592, "a6": 0.1421e-6,
                 "a7": 0., "a8": 0.}  # from J. Appl. Cryst. (1982). 15, 581-589

    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "tof_intensity_incident"

    def __init__(self, **kwargs) -> NoReturn:
        super(TOFIntensityIncident, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_spectrum(self, time):
        """Calculate spectrum for time in microseconds
        
        Two spectra can be calculated:
            (i) Maxwell (by default);
            
            spectrum = A0 + A1*exp[-A2 time**2]/time**5 + A3*exp[-A4 time**2] + 
                A5*exp[-A6 time**3] + A7*exp[-A8 time**4]
            
            (ii) Empirical-Exponents.

            spectrum = A0 + A1*exp[-A2 time] + A3*exp[-A4 time**2] + 
                A5*exp[-A6 time**3] + A7*exp[-A8 time**4]
            
        Parameters
        ----------
        time : TYPE
            DESCRIPTION.

        Returns
        -------
        res : TYPE
            DESCRIPTION.

        """
        exp = numpy.exp
        time_sq = numpy.square(time)
        time_4 = numpy.square(time_sq)
        if self.spectrum == "Empirical-Exponents":
            res = self.a0 + self.a1 * exp(-self.a2 * time) + \
                self.a3 * exp(-self.a4 * time_sq) + \
                self.a5 * exp(-self.a6 * time*time_sq) + \
                self.a7 * exp(-self.a8 * time_4) 
        else:  # self.spectrum == "Maxwell"
            res = self.a0 + self.a1 * exp(-self.a2 * time_sq)/(time*time_4) + \
                self.a3 * exp(-self.a4 * time_sq) + \
                self.a5 * exp(-self.a6 * time*time_sq) + \
                self.a7 * exp(-self.a8 * numpy.square(time_sq)) 
        return res


    def get_coefficients(self):
        coefficients = numpy.array([
            self.a0, self.a1, self.a2, self.a3, 
            self.a4, self.a5, self.a6, self.a7, 
            self.a8], dtype=float)
        return coefficients

    def get_flags_coefficients(self):
        flags_coefficients = numpy.array([
            self.a0_refinement, self.a1_refinement, self.a2_refinement, self.a3_refinement, 
            self.a4_refinement, self.a5_refinement, self.a6_refinement, self.a7_refinement, 
            self.a8_refinement], dtype=bool)
        return flags_coefficients

    def plots(self):
        return [self.plot_spectrum()]
    
    def plot_spectrum(self):
        time_min, time_max = 4000, 18000
        time = numpy.linspace(time_min, time_max, 100)
        spectrum = self.calc_spectrum(time)
        
        fig, ax = plt.subplots()
        ax.set_title("Incident spectrum")
        ax.set_xlabel("time (mc. sec.)")
        ax.set_xlim(time_min, time_max)
        ax.set_ylabel('Incident spectrup')
        ax.plot(time, spectrum, "k-", label=self.spectrum)
        ax.legend(loc='upper right')
        fig.tight_layout()
        return (fig, ax)
    
    def get_dictionary(self):
        res = {}
        res["spectrum_incident_type"] = tof_intensity_incident.spectrum
        res["spectrum_incident_coefficients"] = tof_intensity_incident.get_coefficients()
        res["flags_spectrum_incident_coefficients"] = tof_intensity_incident.get_flags_coefficients()
        return res
        

class TOFIntensityIncidentL(LoopN):
    """Correction of incident intensity for time-of-flight experiment.

    """
    ITEM_CLASS = TOFIntensityIncident
    ATTR_INDEX = None
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(TOFIntensityIncidentL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name
   

# s_cont = """
# _tof_intensity_incident_a1 0
# _tof_intensity_incident_a2 0
# _tof_intensity_incident_a3 0
# _tof_intensity_incident_a4 0
# _tof_intensity_incident_a5 0
# _tof_intensity_incident_a6 0
# _tof_intensity_incident_a7 0
# _tof_intensity_incident_a8 0
# """

# obj = TOFCorrectionIntensityIncident.from_cif(s_cont)
# print(obj, end="\n\n")



from typing import NoReturn
import numpy
import matplotlib.pyplot as plt

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class PdInstrResolution(ItemN):
    """Resolution of Bragg reflections for 1d powder diffractometer.

    Attributes
    ----------
        - u, v, w, x, y (mandatory)

    """
    ATTR_MANDATORY_NAMES = ("u", "v", "w", "x", "y")
    ATTR_MANDATORY_TYPES = (float, float, float, float, float)
    ATTR_MANDATORY_CIF = ("U", "V", "W", "X", "Y")

    ATTR_OPTIONAL_NAMES = ()
    ATTR_OPTIONAL_TYPES = ()
    ATTR_OPTIONAL_CIF = ()

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("u", "v", "w", "x", "y")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"u": "{:.5f}", "v": "{:.5f}", "w": "{:.5f}", "x": "{:.5f}",
                 "y": "{:.5f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {"u": 0., "v": 0., "w": 0., "x": 0., "y": 0.}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "pd_instr_resolution"

    def __init__(self, **kwargs) -> NoReturn:
        super(PdInstrResolution, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_tancos(self, th_hkl):
        """
        tth_hkl in radianas
        calculate tangenth (theta)
        """
        self.t_th = numpy.tan(th_hkl)
        self.t_th_sq = self.t_th**2
        res = numpy.cos(th_hkl)

        self.c_th = res
        self.ic_th = 1./res

    def calc_hg(self, phase_igsize: float = 0., phase_u: float = 0.,
            phase_v: float = 0., phase_w: float = 0., ):
        """
        ttheta in radians, could be array
        gauss size
        """
        u = float(self.u)+phase_u
        v = float(self.v)+phase_v
        w = float(self.w)+phase_w
        res_sq = (u*self.t_th_sq + v*self.t_th + w +
                  phase_igsize*self.ic_th**2)
        self.hg = numpy.sqrt(res_sq)

    def calc_hl(self, phase_x: float = 0., phase_y: float = 0.):
        """
        ttheta in radians, could be array
        lorentz site
        """
        x, y = float(self.x)+phase_x, float(self.y)+phase_y
        self.hl = x*self.t_th + y*self.ic_th

    def calc_hpveta(self):
        """
        ttheta in radians, could be array
        pseudo-Voight function
        """
        hg = self.hg
        hl = self.hl

        hg_2, hl_2 = hg**2, hl**2
        hg_3, hl_3 = hg_2*hg, hl_2*hl
        hg_4, hl_4 = hg_3*hg, hl_3*hl
        hg_5, hl_5 = hg_4*hg, hl_4*hl
        c_2, c_3, c_4, c_5 = 2.69269, 2.42843, 4.47163, 0.07842
        hpv = (hg_5 + c_2*hg_4*hl + c_3*hg_3*hl_2 +
                       c_4*hg_2*hl_3 + c_5*hg*hl_4 + hl_5)**0.2
        hh = hl*1./hpv
        self.hpv = hpv
        self.eta = 1.36603*hh - 0.47719*hh**2 + 0.11116*hh**3

    def calc_agbg(self):
        hpv = self.hpv

        self.ag = (2./hpv)*(numpy.log(2.)/numpy.pi)**0.5
        self.bg = 4*numpy.log(2)/(hpv**2)

    def calc_albl(self):
        hpv = self.hpv
        self.al = 2./(numpy.pi*hpv )
        self.bl = 4./(hpv**2)

    def calc_resolution(
            self, tth_hkl, phase_igsize: float = 0., phase_u: float = 0.,
            phase_v: float = 0., phase_w: float = 0., phase_x: float = 0.,
            phase_y: float = 0.):
        """Calculate parameters for tth tth_hkl in degrees.

        Output
        ------
            - h_pv, eta, h_g, h_l, a_g, b_g, a_l, b_l
        """
        self.calc_tancos(0.5*tth_hkl*numpy.pi/180.)
        self.calc_hg(phase_igsize=phase_igsize, phase_u=phase_u,
                     phase_v=phase_v, phase_w=phase_w)
        self.calc_hl(phase_x=phase_x, phase_y=phase_y)
        self.calc_hpveta()
        self.calc_agbg()
        self.calc_albl()

        a_g = self.ag
        b_g = self.bg
        a_l = self.al
        b_l = self.bl
        h_g = self.hg
        h_l = self.hl
        h_pv = self.hpv
        eta = self.eta

        return h_pv, eta, h_g, h_l, a_g, b_g, a_l, b_l

    def plot_resolution(self):
        """Plot resolution."""
        x_min, x_max = 0, 140
        ttheta = numpy.linspace(x_min, x_max, 100)
        h_pv, eta, h_g, h_l, a_g, b_g, a_l, b_l = self.calc_resolution(ttheta)
        fig, ax = plt.subplots()
        ax.set_title("Resolution")
        ax.set_xlabel("2 theta (degrees)")
        ax.set_xlim(x_min, x_max)
        ax.set_ylabel('Resolution (degrees)')
        ax.plot(ttheta, h_g, "k:", label="h_g")
        ax.plot(ttheta, h_l, "k--", label="h_l")
        ax.plot(ttheta, h_pv, "k-", label="h_pv")
        ax.legend(loc='upper right')
        fig.tight_layout()
        return (fig, ax)

    def plots(self):
        return [self.plot_resolution()]

    def get_dictionary(self):
        res = {}
        res["resolution_parameters"] = numpy.array([
            self.u, self.v, self.w,
            self.x, self.y], dtype=float)

        res["flags_resolution_parameters"] = numpy.array([
            self.u_refinement, self.v_refinement, self.w_refinement,
            self.x_refinement, self.y_refinement], dtype=bool)
        return res


class PdInstrResolutionL(LoopN):
    """Resolution of Bragg reflections for several 1d powder diffractometers.

    """
    ITEM_CLASS = PdInstrResolution
    ATTR_INDEX = None
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(PdInstrResolutionL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name


# s_cont = """
#  loop_

#  _pd_instr_resolution_u
#  _pd_instr_resolution_v
#  _pd_instr_resolution_w
#  _pd_instr_resolution_x
#  _pd_instr_resolution_y
# 16.9776(25) -2.8357 0.5763 0. 0.
# """

# obj = PdInstrResolutionL.from_cif(s_cont)
# print(obj, end="\n\n")
